"""
Django Ledger created by Miguel Sanda <msanda@arrobalytics.com>.
Copyright© EDMA Group Inc licensed under the GPLv3 Agreement.

A purchase order is a commercial source document that is issued by a business purchasing department when placing an
order with its vendors or suppliers. The document indicates the details on the items that are to be purchased, such as
the types of goods, quantity, and price. In simple terms, it is the contract drafted by the buyer when purchasing goods
from the seller.

The PurchaseOrderModel is designed to track the status of a Purchase Order and all its items. The PurchaseOrderModel
starts in draft model by default and goes through different states including InReview, Approved, Fulfilled, Canceled and
Void. The PurchaseOrderModel also keeps track of when these states take place.

"""
import warnings
from datetime import date
from string import ascii_uppercase, digits
from typing import Tuple, List, Union, Optional, Dict
from uuid import uuid4, UUID

from django.contrib.auth import get_user_model
from django.core.exceptions import ValidationError, ObjectDoesNotExist
from django.core.validators import MinLengthValidator
from django.db import models, transaction, IntegrityError
from django.db.models import Q, Sum, Count, F, Manager, QuerySet
from django.db.models.functions import Coalesce
from django.db.models.signals import pre_save
from django.shortcuts import get_object_or_404
from django.urls import reverse
from django.utils.translation import gettext_lazy as _

from django_ledger.io.io_core import get_localdate
from django_ledger.models.bill import BillModel, BillModelQuerySet
from django_ledger.models.deprecations import deprecated_entity_slug_behavior
from django_ledger.models.entity import EntityModel
from django_ledger.models.items import ItemTransactionModel, ItemTransactionModelQuerySet, ItemModelQuerySet, ItemModel
from django_ledger.models.mixins import CreateUpdateMixIn, MarkdownNotesMixIn, ItemizeMixIn
from django_ledger.models.signals import (
    po_status_draft,
    po_status_void,
    po_status_fulfilled,
    po_status_approved,
    po_status_canceled,
    po_status_in_review
)
from django_ledger.models.utils import lazy_loader
from django_ledger.settings import DJANGO_LEDGER_DOCUMENT_NUMBER_PADDING, DJANGO_LEDGER_PO_NUMBER_PREFIX, \
    DJANGO_LEDGER_USE_DEPRECATED_BEHAVIOR

PO_NUMBER_CHARS = ascii_uppercase + digits

UserModel = get_user_model()


class PurchaseOrderModelValidationError(ValidationError):
    pass


class PurchaseOrderModelQuerySet(QuerySet):
    """
    A custom-defined PurchaseOrderModel QuerySet.
    """

    def for_user(self, user_model):
        if user_model.is_superuser:
            return self
        return self.filter(
            Q(entity__admin=user_model) |
            Q(entity__managers__in=[user_model])
        )

    def approved(self):
        """
        Filters the QuerySet to include Approved PurchaseOrderModels only.

        Returns
        -------
        PurchaseOrderModelQuerySet
            A PurchaseOrderModelQuerySet with applied filters.
        """
        return self.filter(po_status__exact=PurchaseOrderModel.PO_STATUS_APPROVED)

    def fulfilled(self):
        """
        Filters the QuerySet to include Fulfilled PurchaseOrderModels only.

        Returns
        -------
        PurchaseOrderModelQuerySet
            A PurchaseOrderModelQuerySet with applied filters.
        """
        return self.filter(po_status__exact=PurchaseOrderModel.PO_STATUS_FULFILLED)

    def active(self):
        """
        Filters the QuerySet to include Active PurchaseOrderModels only.
        Active PurchaseOrderModels are either approved or fulfilled, which are those that may contain associated
        transactions on the Ledger.

        Returns
        -------
        PurchaseOrderModelQuerySet
            A PurchaseOrderModelQuerySet with applied filters.
        """
        return self.filter(
            Q(po_status__exact=PurchaseOrderModel.PO_STATUS_APPROVED) |
            Q(po_status__exact=PurchaseOrderModel.PO_STATUS_FULFILLED)
        )

    def draft(self):
        return self.filter(po_status__exact=PurchaseOrderModel.PO_STATUS_DRAFT)


class PurchaseOrderModelManager(Manager):
    """
    A custom defined PurchaseOrderModel Manager.
    """

    @deprecated_entity_slug_behavior
    def for_entity(self, entity_model: EntityModel | str | UUID = None, **kwargs) -> PurchaseOrderModelQuerySet:
        """
        Fetches a QuerySet of PurchaseOrderModel associated with a specific EntityModel & UserModel.
        May pass an instance of EntityModel or a String representing the EntityModel slug.

        Returns
        -------
        PurchaseOrderModelQuerySet
            A PurchaseOrderModelQuerySet with applied filters.
        """

        qs = self.get_queryset()
        if 'user_model' in kwargs:
            warnings.warn(
                'user_model parameter is deprecated and will be removed in a future release. '
                'Use for_user(user_model).for_entity(entity_model) instead to keep current behavior.',
                DeprecationWarning,
                stacklevel=2
            )
            if DJANGO_LEDGER_USE_DEPRECATED_BEHAVIOR:
                qs = qs.for_user(kwargs['user_model'])

        if isinstance(entity_model, EntityModel):
            qs = qs.filter(entity=entity_model)
        elif isinstance(entity_model, str):
            qs = qs.filter(entity__slug__exact=entity_model)
        elif isinstance(entity_model, UUID):
            qs = qs.filter(entity_id=entity_model)
        else:
            raise PurchaseOrderModelValidationError(
                message='Entity slug must be either an EntityModel or a String representing the EntityModel slug',
            )
        return qs


class PurchaseOrderModelAbstract(CreateUpdateMixIn,
                                 ItemizeMixIn,
                                 MarkdownNotesMixIn):
    """
    The base implementation of the PurchaseOrderModel.

    Attributes
    ----------
    uuid: UUID
        This is a unique primary key generated for the table. The default value of this field is uuid4().
    po_number: str
        A unique human-readable and sequential PO Number identifier. Automatically generated.
    po_title: str
        The PurchaseOrderModel instance title.
    po_status: str
        One of PO_STATUS values representing the current status of the PurchaseOrderModel instance.
    po_amount: Decimal
        The total value of the PurchaseOrderModel instance.
    po_amount_received: Decimal
        The PurchaseOrderModel instance total value received to date. Cannot be greater than PO amount.
    entity: EntityModel
        The EntityModel associated with the PurchaseOrderModel instance.
    date_draft: date
        The draft date represents the date when the PurchaseOrderModel was first created. Defaults to
        :func:`localdate <django.utils.timezone.localdate>`.
    date_in_review: date
        The in review date represents the date when the PurchaseOrderModel was marked as In Review status.
        Will be null if PurchaseOrderModel is canceled during draft status. Defaults to
        :func:`localdate <django.utils.timezone.localdate>`.
    date_approved: date
        The approved date represents the date when the PurchaseOrderModel was approved. Will be null if
        PurchaseOrderModel is canceled.
        Defaults to :func:`localdate <django.utils.timezone.localdate>`.
    date_fulfilled: date
        The paid date represents the date when the PurchaseOrderModel was fulfilled and po_amount_received equals
        po_amount. Will be null if PurchaseOrderModel is canceled.
        Defaults to :func:`localdate <django.utils.timezone.localdate>`.
    date_void: date
        The void date represents the date when the PurchaseOrderModel was void, if applicable.
        Will be null unless PurchaseOrderModel is void.
        Defaults to :func:`localdate <django.utils.timezone.localdate>`.
    date_canceled: date
        The canceled date represents the date when the PurchaseOrderModel was canceled, if applicable.
        Will be null unless PurchaseOrderModel is canceled.
        Defaults to :func:`localdate <django.utils.timezone.localdate>`.
    po_items:
        A foreign key reference to the list of ItemTransactionModel that make the PurchaseOrderModel amount.
    ce_model: EstimateModel
        A foreign key reference to the EstimateModel associated with the PurchaseOrderModel, if any.
    """
    PO_STATUS_DRAFT = 'draft'
    PO_STATUS_REVIEW = 'in_review'
    PO_STATUS_APPROVED = 'approved'
    PO_STATUS_FULFILLED = 'fulfilled'
    PO_STATUS_VOID = 'void'
    PO_STATUS_CANCELED = 'canceled'

    """The different valid PO Status and their representation in the Database"""
    PO_STATUS = [
        (PO_STATUS_DRAFT, _('Draft')),
        (PO_STATUS_REVIEW, _('In Review')),
        (PO_STATUS_APPROVED, _('Approved')),
        (PO_STATUS_FULFILLED, _('Fulfilled')),
        (PO_STATUS_CANCELED, _('Canceled')),
        (PO_STATUS_VOID, _('Void')),
    ]

    uuid = models.UUIDField(default=uuid4, editable=False, primary_key=True)
    po_number = models.SlugField(max_length=20, verbose_name=_('Purchase Order Number'), editable=False)
    po_title = models.CharField(max_length=250,
                                verbose_name=_('Purchase Order Title'),
                                validators=[
                                    MinLengthValidator(limit_value=5,
                                                       message=_(
                                                           f'PO Title must be greater than 5'))
                                ])
    po_status = models.CharField(max_length=10, choices=PO_STATUS, default=PO_STATUS[0][0])
    po_amount = models.DecimalField(default=0, decimal_places=2, max_digits=20, verbose_name=_('Purchase Order Amount'))
    po_amount_received = models.DecimalField(default=0,
                                             decimal_places=2,
                                             max_digits=20,
                                             verbose_name=_('Received Amount'))
    entity = models.ForeignKey('django_ledger.EntityModel',
                               on_delete=models.CASCADE,
                               verbose_name=_('Entity'))

    date_draft = models.DateField(null=True, blank=True, verbose_name=_('Draft Date'))
    date_in_review = models.DateField(null=True, blank=True, verbose_name=_('In Review Date'))
    date_approved = models.DateField(null=True, blank=True, verbose_name=_('Approved Date'))
    date_void = models.DateField(blank=True, null=True, verbose_name=_('Void Date'))
    date_fulfilled = models.DateField(blank=True, null=True, verbose_name=_('Fulfillment Date'))
    date_canceled = models.DateField(null=True, blank=True, verbose_name=_('Canceled Date'))

    po_items = models.ManyToManyField('django_ledger.ItemModel',
                                      through='django_ledger.ItemTransactionModel',
                                      through_fields=('po_model', 'item_model'),
                                      verbose_name=_('Purchase Order Items'))

    ce_model = models.ForeignKey('django_ledger.EstimateModel',
                                 on_delete=models.RESTRICT,
                                 null=True,
                                 blank=True,
                                 verbose_name=_('Associated Customer Job/Estimate'))

    objects = PurchaseOrderModelManager.from_queryset(queryset_class=PurchaseOrderModelQuerySet)()

    class Meta:
        abstract = True
        indexes = [
            models.Index(fields=['entity']),
            models.Index(fields=['po_number']),
            models.Index(fields=['po_status']),
            models.Index(fields=['ce_model']),

            models.Index(fields=['date_draft']),
            models.Index(fields=['date_in_review']),
            models.Index(fields=['date_approved']),
            models.Index(fields=['date_fulfilled']),
            models.Index(fields=['date_canceled']),
            models.Index(fields=['date_void']),
        ]
        unique_together = [
            ('entity', 'po_number')
        ]

    def __str__(self):
        # pylint: disable=no-member
        return f'PO Model: {self.po_number} | {self.get_po_status_display()}'

    # Configuration...

    def is_configured(self) -> bool:
        return all([
            self.entity_id is not None,
            self.date_draft,
            self.po_status
        ])

    def configure(self,
                  entity_slug: Union[str, EntityModel],
                  po_title: Optional[str] = None,
                  user_model: Optional[UserModel] = None,
                  draft_date: Optional[date] = None,
                  estimate_model=None,
                  commit: bool = False):
        """
        A configuration hook which executes all initial PurchaseOrderModel setup on to the EntityModel and all initial
        values of the EntityModel. Can only call this method once in the lifetime of a PurchaseOrderModel.

        Parameters
        __________
        entity_slug: str or EntityModel
            The entity slug or EntityModel to associate the Bill with.
        user_model:
            The UserModel making the request to check for QuerySet permissions.
        ledger_posted:
            An option to mark the BillModel Ledger as posted at the time of configuration. Defaults to False.
        bill_desc: str
            An optional description appended to the LedgerModel name.
        commit: bool
            Saves the current BillModel after being configured.

        Returns
        -------
        PurchaseOrderModel
            The configured PurchaseOrderModel instance.
        """
        if not self.is_configured():
            if isinstance(entity_slug, str):
                if not user_model:
                    raise PurchaseOrderModelValidationError(_('Must pass user_model when using entity_slug.'))
                entity_qs = EntityModel.objects.for_user(user_model=user_model)
                entity_model: EntityModel = get_object_or_404(entity_qs, slug__exact=entity_slug)
            elif isinstance(entity_slug, EntityModel):
                entity_model = entity_slug
            else:
                raise PurchaseOrderModelValidationError('entity_slug must be an instance of str or EntityModel')

            self.date_draft = get_localdate() if not draft_date else draft_date
            self.po_status = PurchaseOrderModel.PO_STATUS_DRAFT

            if estimate_model:
                self.action_bind_estimate(estimate_model=estimate_model, commit=False)

            self.entity = entity_model

            if self.can_generate_po_number():
                self.generate_po_number(commit=commit)

            if not self.po_title:
                if all([
                    po_title is None,
                    self.po_title is None or not len(self.po_title),
                    self.po_number is not None
                ]):
                    self.po_title = f'PO Number {self.po_number}'
                else:
                    self.po_title = po_title

            self.clean_fields()
            self.clean()
            if commit:
                self.save()
        return self

    def validate_item_transaction_qs(self, queryset: Union[ItemTransactionModelQuerySet, List[ItemTransactionModel]]):
        """
        Validates that the entire ItemTransactionModelQuerySet is bound to the PurchaseOrderModel.

        Parameters
        ----------
        queryset: ItemTransactionModelQuerySet or list of ItemTransactionModel.
            ItemTransactionModelQuerySet to validate.
        """
        valid = all([
            i.po_model_id == self.uuid for i in queryset
        ])
        if not valid:
            raise PurchaseOrderModelValidationError(f'Invalid queryset. All items must be assigned to PO {self.uuid}')

    # ### ItemizeMixIn implementation START...

    def can_migrate_itemtxs(self) -> bool:
        return self.is_draft()

    def migrate_itemtxs(self, itemtxs: Dict, operation: str, commit: bool = False):
        itemtxs_batch = super().migrate_itemtxs(itemtxs=itemtxs, commit=commit, operation=operation)
        self.update_state(itemtxs_qs=itemtxs_batch)
        self.clean()
        if commit:
            self.save(update_fields=['po_amount',
                                     'po_amount_received',
                                     'updated'])
        return itemtxs_batch

    def get_item_model_qs(self) -> ItemModelQuerySet:
        return ItemModel.objects.filter(
            entity_id__exact=self.entity_id
        ).purchase_orders()

    def get_itemtxs_data(self,
                         queryset: Optional[Union[ItemTransactionModelQuerySet, List[ItemTransactionModel]]] = None,
                         aggregate_on_db: bool = False,
                         lazy_agg: bool = False) -> Tuple:
        """
        Fetches the PurchaseOrderModel Items and aggregates the QuerySet.

        Parameters
        ----------
        queryset: ItemTransactionModelQuerySet
            Optional pre-fetched ItemModelQueryset to use. Avoids additional DB query if provided.
            Validated if provided.
        aggregate_on_db: bool
            If True, performs aggregation of ItemsTransactions in the DB resulting in one additional DB query.

        Returns
        -------
        A tuple: ItemTransactionModelQuerySet, dict
        """
        if not queryset:
            queryset = self.itemtransactionmodel_set.all().select_related('bill_model', 'item_model')
        else:
            self.validate_item_transaction_qs(queryset)

        if aggregate_on_db and isinstance(queryset, ItemTransactionModelQuerySet):
            return queryset, queryset.aggregate(
                po_total_amount__sum=Coalesce(Sum('po_total_amount'), 0.0, output_field=models.FloatField()),
                bill_amount_paid__sum=Coalesce(Sum('bill_model__amount_paid'), 0.0, output_field=models.FloatField()),
                total_items=Count('uuid')
            )
        return queryset, {
            'po_total_amount__sum': sum(i.total_amount for i in queryset),
            'bill_amount_paid__sum': sum(i.bill_model.amount_paid for i in queryset if i.bill_model_id),
            'total_items': len(queryset)
        } if not lazy_agg else None

    # ### ItemizeMixIn implementation END...
    def update_state(self, itemtxs_qs: Optional[
        Union[ItemTransactionModelQuerySet, List[ItemTransactionModel]]] = None) -> Tuple:

        """
        Updates the state of the PurchaseOrderModel.

        Parameters
        ----------
        itemtxs_qs: ItemTransactionModelQuerySet or list of ItemTransactionModel

        Returns
        -------
        tuple
            A tuple of ItemTransactionModels and Aggregation
        """
        itemtxs_qs, itemtxs_agg = self.get_itemtxs_data(queryset=itemtxs_qs)

        if isinstance(itemtxs_qs, list):
            self.po_amount = round(sum(a.po_total_amount for a in itemtxs_qs if not a.is_canceled()), 2)
            self.po_amount_received = round(sum(a.po_total_amount for a in itemtxs_qs if a.is_received()), 2)
        elif isinstance(itemtxs_qs, ItemTransactionModelQuerySet):
            total_po_amount = round(sum(i.po_total_amount for i in itemtxs_qs if not i.is_canceled()), 2)
            total_received = round(sum(i.po_total_amount for i in itemtxs_qs if i.is_received()), 2)
            self.po_amount = total_po_amount
            self.po_amount_received = total_received

        return itemtxs_qs, itemtxs_agg

    # State...
    def is_draft(self) -> bool:
        """
        Checks if the PurchaseOrderModel is in Draft status.

        Returns
        -------
        bool
            True if PurchaseOrderModel is Draft, else False.
        """
        return self.po_status == self.PO_STATUS_DRAFT

    def is_review(self) -> bool:
        """
        Checks if the PurchaseOrderModel is in Review status.

        Returns
        -------
        bool
            True if PurchaseOrderModel is Review, else False.
        """
        return self.po_status == self.PO_STATUS_REVIEW

    def is_approved(self) -> bool:
        """
        Checks if the PurchaseOrderModel is in Approved status.

        Returns
        -------
        bool
            True if PurchaseOrderModel is Approved, else False.
        """
        return self.po_status == self.PO_STATUS_APPROVED

    def is_fulfilled(self) -> bool:
        """
        Checks if the PurchaseOrderModel is in Fulfilled status.

        Returns
        -------
        bool
            True if PurchaseOrderModel is in Fulfilled status, else False.
        """
        return self.po_status == self.PO_STATUS_FULFILLED

    def is_canceled(self) -> bool:
        """
        Checks if the PurchaseOrderModel is in Canceled status.

        Returns
        -------
        bool
            True if PurchaseOrderModel is in Canceled, else False.
        """
        return self.po_status == self.PO_STATUS_CANCELED

    def is_void(self) -> bool:
        """
        Checks if the PurchaseOrderModel is in Void status.

        Returns
        -------
        bool
            True if PurchaseOrderModel is Void, else False.
        """
        return self.po_status == self.PO_STATUS_VOID

    # Permissions...
    def can_draft(self) -> bool:
        """
        Checks if the PurchaseOrderModel can be marked as Draft.

        Returns
        -------
        bool
            True if PurchaseOrderModel can be marked as Draft, else False.
        """
        return self.is_review()

    def can_review(self) -> bool:
        """
        Checks if the PurchaseOrderModel can be marked as In Review.

        Returns
        -------
        bool
            True if PurchaseOrderModel can be marked as In Review, else False.
        """
        return self.is_draft()

    def can_approve(self) -> bool:
        """
        Checks if the PurchaseOrderModel can be marked as Approved.

        Returns
        -------
        bool
            True if PurchaseOrderModel can be marked as Approved, else False.
        """
        return self.is_review()

    def can_fulfill(self) -> bool:
        """
        Checks if the PurchaseOrderModel can be marked as Fulfilled.

        Returns
        -------
        bool
            True if PurchaseOrderModel can be marked as Fulfilled, else False.
        """
        return self.is_approved()

    def can_cancel(self) -> bool:
        """
        Checks if the PurchaseOrderModel can be marked as Canceled.

        Returns
        -------
        bool
            True if PurchaseOrderModel can be marked as Canceled, else False.
        """
        return any([
            self.is_draft(),
            self.is_review()
        ])

    def can_void(self) -> bool:
        """
        Checks if the PurchaseOrderModel can be marked as Void.

        Returns
        -------
        bool
            True if PurchaseOrderModel can be marked as Void, else False.
        """
        return self.is_approved()

    def can_delete(self) -> bool:
        """
        Checks if the PurchaseOrderModel can be deleted.

        Returns
        -------
        bool
            True if PurchaseOrderModel can be deleted, else False.
        """
        return any([
            self.is_draft(),
            self.is_review()
        ])

    def can_edit_items(self) -> bool:
        """
        Checks if the PurchaseOrderModel items can be edited.

        Returns
        -------
        bool
            True if PurchaseOrderModel items can be edited, else False.
        """
        return self.is_draft()

    def is_contract_bound(self):
        """
        Checks if the PurchaseOrderModel is bound to an EstimateModel.

        Returns
        -------
        bool
            True if PurchaseOrderModel is bound to an EstimateModel, else False.
        """
        return self.ce_model_id is not None

    def can_bind_estimate(self, estimate_model, raise_exception: bool = False) -> bool:
        """
        Checks if the PurchaseOrderModel ican be bound to an EstimateModel.

        Returns
        -------
        bool
            True if PurchaseOrderModel can be bound to an EstimateModel, else False.
        """
        if self.is_contract_bound():
            if raise_exception:
                raise PurchaseOrderModelValidationError(
                    f'PO {self.po_number} already bound to Estimate {self.ce_model.estimate_number}')
            return False
        elif self.entity_id != estimate_model.entity_id:
            if raise_exception:
                raise PurchaseOrderModelValidationError(
                    f'Invalid EstimateModel for entity {self.entity.slug}'
                )
            return False

        # check if estimate_model is passed and raise exception if needed...
        is_approved = estimate_model.is_approved()
        if not is_approved and raise_exception:
            raise PurchaseOrderModelValidationError(f'Cannot bind estimate that is not approved.')
        return all([
            is_approved
        ])

    def action_bind_estimate(self, estimate_model, commit: bool = False):
        """
        Binds a specific EstimateModel to the PurchaseOrderModel instance.

        Parameters
        ----------
        estimate_model: EstimateModel
            The EstimateModel to bind.
        commit: bool
            Commits the changes in the Database, if True. Defaults to False.
        """
        try:
            self.can_bind_estimate(estimate_model, raise_exception=True)
        except ValueError as e:
            raise e
        self.ce_model = estimate_model
        self.clean()
        if commit:
            self.save(update_fields=[
                'ce_model',
                'updated'
            ])

    def can_generate_po_number(self):
        """
        Checks if PurchaseOrderModel can generate its Document Number.

        Returns
        -------
        bool
            True if PurchaseOrderModel can generate its po_number, else False.
        """
        return all([
            self.date_draft,
            not self.po_number
        ])

    # Actions...

    # DRAFT...
    def mark_as_draft(self, date_draft: Optional[date] = None, commit: bool = False, **kwargs):
        """
        Marks PurchaseOrderModel as Draft.

        Parameters
        ----------
        date_draft: date
            Draft date. If None, defaults to localdate().
        commit: bool
            Commits transaction into the Database. Defaults to False.
        """
        if not self.can_draft():
            raise PurchaseOrderModelValidationError(
                message=f'Purchase Order {self.po_number} cannot be marked as draft.')
        self.po_status = self.PO_STATUS_DRAFT
        self.date_draft = get_localdate() if not date_draft else date_draft
        self.clean()
        if commit:
            self.save(update_fields=[
                'po_status',
                'updated'
            ])
        po_status_draft.send_robust(sender=self.__class__,
                                    instance=self,
                                    commited=commit,
                                    **kwargs)

    def get_mark_as_draft_html_id(self):
        """
        PurchaseOrderModel Mark as Draft HTML ID Tag.

        Returns
        -------
        str
            HTML ID as a String.
        """
        return f'djl-{self.uuid}-po-mark-as-draft'

    def get_mark_as_draft_url(self):
        """
        PurchaseOrderModel Mark as Draft URL ID Tag.

        Returns
        -------
        str
            HTML ID as a String.
        """
        return reverse('django_ledger:po-action-mark-as-draft',
                       kwargs={
                           'entity_slug': self.entity.slug,
                           'po_pk': self.uuid
                       })

    def get_mark_as_draft_message(self):
        """
        PurchaseOrderModel Mark as Draft Message.

        Returns
        -------
        str
            Message as a String.
        """
        return _('Do you want to mark Purchase Order %s as Draft?') % self.po_number

    # REVIEW...
    def mark_as_review(self, date_in_review: Optional[date] = None, commit: bool = False, **kwargs):
        """
        Marks PurchaseOrderModel as In Review.

        Parameters
        ----------
        date_in_review: date
            Draft date. If None, defaults to localdate().
        commit: bool
            Commits transaction into the Database. Defaults to False.
        """
        if not self.can_review():
            raise PurchaseOrderModelValidationError(
                message=f'Purchase Order {self.po_number} cannot be marked as in review.')

        itemtxs_qs, itemtxs_agg = self.get_itemtxs_data()
        if not itemtxs_qs.count():
            raise PurchaseOrderModelValidationError(message='Cannot review a PO without items...')
        if not self.po_amount:
            raise PurchaseOrderModelValidationError(message='PO amount is zero.')

        self.date_in_review = get_localdate() if not date_in_review else date_in_review
        self.po_status = self.PO_STATUS_REVIEW
        self.clean()
        if commit:
            self.save(update_fields=[
                'po_status',
                'date_in_review',
                'updated'
            ])
        po_status_in_review.send_robust(sender=self.__class__,
                                        instance=self,
                                        commited=commit,
                                        **kwargs)

    def get_mark_as_review_html_id(self):
        """
        PurchaseOrderModel Mark as In Review HTML ID Tag.

        Returns
        -------
        str
            HTML ID as a String.
        """
        return f'djl-{self.uuid}-po-mark-as-review'

    def get_mark_as_review_url(self):
        """
        PurchaseOrderModel Mark as In Review URL ID Tag.

        Returns
        -------
        str
            HTML ID as a String.
        """
        return reverse('django_ledger:po-action-mark-as-review',
                       kwargs={
                           'entity_slug': self.entity.slug,
                           'po_pk': self.uuid
                       })

    def get_mark_as_review_message(self):
        """
        PurchaseOrderModel Mark as Review Message.

        Returns
        -------
        str
            Message as a String.
        """
        return _('Do you want to mark Purchase Order %s as In Review?') % self.po_number

    # APPROVED...
    def mark_as_approved(self, date_approved: Optional[date] = None, commit: bool = False, **kwargs):
        """
        Marks PurchaseOrderModel as Approved.

        Parameters
        ----------
        date_approved: date
            Approved date. If None, defaults to localdate().
        commit: bool
            Commits transaction into the Database. Defaults to False.
        """
        if not self.can_approve():
            raise PurchaseOrderModelValidationError(
                message=f'Purchase Order {self.po_number} cannot be marked as approved.')
        self.date_approved = get_localdate() if not date_approved else date_approved
        self.po_status = self.PO_STATUS_APPROVED
        self.clean()
        if commit:
            self.itemtransactionmodel_set.all().update(po_item_status=ItemTransactionModel.STATUS_NOT_ORDERED)
            self.save(update_fields=[
                'date_approved',
                'po_status',
                'updated'
            ])
        po_status_approved.send_robust(sender=self.__class__,
                                       instance=self,
                                       commited=commit,
                                       **kwargs)

    def get_mark_as_approved_html_id(self):
        """
        PurchaseOrderModel Mark as Approved HTML ID Tag.

        Returns
        -------
        str
            HTML ID as a String.
        """
        return f'djl-{self.uuid}-po-mark-as-approved'

    def get_mark_as_approved_url(self):
        """
        PurchaseOrderModel Mark as Approved URL ID Tag.

        Returns
        -------
        str
            HTML ID as a String.
        """
        return reverse('django_ledger:po-action-mark-as-approved',
                       kwargs={
                           'entity_slug': self.entity.slug,
                           'po_pk': self.uuid
                       })

    def get_mark_as_approved_message(self):
        """
        PurchaseOrderModel Mark as Approved Message.

        Returns
        -------
        str
            Message as a String.
        """
        return _('Do you want to mark Purchase Order %s as Approved?') % self.po_number

    # CANCEL...
    def mark_as_canceled(self, date_canceled: Optional[date] = None, commit: bool = False, **kwargs):
        """
        Marks PurchaseOrderModel as Canceled.

        Parameters
        ----------
        date_canceled: date
            Canceled date. If None, defaults to localdate().
        commit: bool
            Commits transaction into the Database. Defaults to False.
        """
        if not self.can_cancel():
            raise PurchaseOrderModelValidationError(
                message=f'Purchase Order {self.po_number} cannot be marked as canceled.')
        self.date_canceled = get_localdate() if not date_canceled else date_canceled
        self.po_status = self.PO_STATUS_CANCELED
        self.clean()
        if commit:
            self.save(update_fields=[
                'po_status',
                'date_canceled',
                'updated'
            ])
        po_status_canceled.send_robust(sender=self.__class__,
                                       instance=self,
                                       commited=commit,
                                       **kwargs)

    def get_mark_as_canceled_html_id(self):
        """
        PurchaseOrderModel Mark as Canceled HTML ID Tag.

        Returns
        -------
        str
            HTML ID as a String.
        """
        return f'djl-{self.uuid}-po-mark-as-canceled'

    def get_mark_as_canceled_url(self):
        """
        PurchaseOrderModel Mark as Canceled URL ID Tag.

        Returns
        -------
        str
            HTML ID as a String.
        """
        return reverse('django_ledger:po-action-mark-as-canceled',
                       kwargs={
                           'entity_slug': self.entity.slug,
                           'po_pk': self.uuid
                       })

    def get_mark_as_canceled_message(self):
        """
        PurchaseOrderModel Mark as Canceled Message.

        Returns
        -------
        str
            Message as a String.
        """
        return _('Do you want to mark Purchase Order %s as Canceled?') % self.po_number

    # FULFILL...
    def mark_as_fulfilled(self,
                          date_fulfilled: date = None,
                          po_items: Union[ItemTransactionModelQuerySet, List[ItemTransactionModel]] = None,
                          commit: bool = False,
                          **kwargs):
        """
        Marks PurchaseOrderModel as Fulfilled.

        Parameters
        ----------
        date_fulfilled: date
            Fulfilled date. If None, defaults to localdate().
        po_items: ItemTransactionModelQuerySet or list of ItemTransactionModel.
            Pre-fetched ItemTransactionModelQuerySet or list of  ItemTransactionModel.
            Validated if provided.
        commit: bool
            Commits transaction into the Database. Defaults to False.
        """

        if not self.can_fulfill():
            raise PurchaseOrderModelValidationError(
                message=f'Purchase Order {self.po_number} cannot be marked as fulfilled.')

        if not po_items:
            po_items, po_items_agg = self.get_itemtxs_data(queryset=po_items)

        self.date_fulfilled = get_localdate() if not date_fulfilled else date_fulfilled
        self.po_amount_received = self.po_amount

        bill_models = [i.bill_model for i in po_items]
        all_items_billed = all(bill_models)
        if not all_items_billed:
            raise PurchaseOrderModelValidationError('All items must be billed before PO can be fulfilled.')

        all_bills_paid = all(b.is_paid() for b in bill_models)
        if not all_bills_paid:
            raise PurchaseOrderModelValidationError('All Bills must be paid before PO can be fulfilled.')

        all_items_received = all(i.is_received() for i in po_items)
        if not all_items_received:
            raise PurchaseOrderModelValidationError('All items must be received before PO is fulfilled.')

        self.date_fulfilled = date_fulfilled
        self.po_status = self.PO_STATUS_FULFILLED
        self.clean()

        if commit:
            # todo: what if PO items is list???...
            po_items.update(po_item_status=ItemTransactionModel.STATUS_RECEIVED)
            self.save(update_fields=[
                'date_fulfilled',
                'po_status',
                'updated'
            ])
        po_status_fulfilled.send_robust(sender=self.__class__,
                                        instance=self,
                                        commited=commit,
                                        **kwargs)

    def get_mark_as_fulfilled_html_id(self):
        """
        PurchaseOrderModel Mark as Fulfilled HTML ID Tag.

        Returns
        -------
        str
            HTML ID as a String.
        """
        return f'djl-{self.uuid}-po-mark-as-fulfilled'

    def get_mark_as_fulfilled_url(self):
        """
        PurchaseOrderModel Mark as Fulfilled URL ID Tag.

        Returns
        -------
        str
            URL as a String.
        """
        return reverse('django_ledger:po-action-mark-as-fulfilled',
                       kwargs={
                           'entity_slug': self.entity.slug,
                           'po_pk': self.uuid
                       })

    def get_mark_as_fulfilled_message(self):
        """
        PurchaseOrderModel Mark as Fulfilled Message.

        Returns
        -------
        str
            Message as a String.
        """
        return _('Do you want to mark Purchase Order %s as Fulfilled?') % self.po_number

    # VOID...
    def mark_as_void(self,
                     void_date: Optional[date] = None,
                     commit: bool = False,
                     **kwargs):
        """
        Marks PurchaseOrderModel as Fulfilled.

        Parameters
        ----------
        void_date: date
            Void date. If None, defaults to localdate().
        commit: bool
            Commits transaction into the Database. Defaults to False.
        """
        if not self.can_void():
            raise PurchaseOrderModelValidationError(
                message=f'Purchase Order {self.po_number} cannot be marked as void.')

        # all bills associated with this PO...
        bill_model_qs = self.get_po_bill_queryset()
        bill_model_qs = bill_model_qs.only('bill_status')

        if not all(b.is_void() for b in bill_model_qs):
            raise PurchaseOrderModelValidationError('Must void all PO bills before PO can be voided.')

        self.date_void = get_localdate() if not void_date else void_date
        self.po_status = self.PO_STATUS_VOID
        self.clean()

        if commit:
            self.save(update_fields=[
                'void_date',
                'po_status',
                'updated'
            ])
        po_status_void.send_robust(sender=self.__class__,
                                   instance=self,
                                   commited=commit,
                                   **kwargs)

    def get_mark_as_void_html_id(self):
        """
        PurchaseOrderModel Mark as Void HTML ID Tag.

        Returns
        -------
        str
            HTML ID as a String.
        """
        return f'djl-{self.uuid}-po-mark-as-void'

    def get_mark_as_void_url(self):
        """
        PurchaseOrderModel Mark as Void URL ID Tag.

        Returns
        -------
        str
            HTML ID as a String.
        """
        return reverse('django_ledger:po-action-mark-as-void',
                       kwargs={
                           'entity_slug': self.entity.slug,
                           'po_pk': self.uuid
                       })

    def get_mark_as_void_message(self):
        """
        PurchaseOrderModel Mark as Void Message.

        Returns
        -------
        str
            Message as a String.
        """
        return _('Do you want to mark Purchase Order %s as Void?') % self.po_number

    def get_po_bill_queryset(self) -> BillModelQuerySet:
        """
        Fetches a BillModelQuerySet of all BillModels associated with the PurchaseOrderModel instance.

        Returns
        -------
        BillModelQuerySet
        """
        return BillModel.objects.filter(bill_items__purchaseordermodel__uuid__exact=self.uuid)

    def get_status_action_date(self):
        """
        Current status action date.

        Returns
        -------
        date
            A date. i.e. If status is Approved, return date_approved. If In Review, return date_in_review.
        """
        return getattr(self, f'date_{self.po_status}')

    def _get_next_state_model(self, raise_exception: bool = True):
        """
        Fetches the next sequenced state model associated with the PurchaseOrderModel number.

        Parameters
        ----------
        raise_exception: bool
            Raises IntegrityError if unable to secure transaction from DB.

        Returns
        -------
        EntityStateModel
            An instance of EntityStateModel
        """
        EntityStateModel = lazy_loader.get_entity_state_model()
        EntityModel = lazy_loader.get_entity_model()
        entity_model = EntityModel.objects.get(uuid__exact=self.entity_id)
        fy_key = entity_model.get_fy_for_date(dt=self.date_draft)
        try:
            LOOKUP = {
                'entity_model_id__exact': self.entity_id,
                'entity_unit_id': None,
                'fiscal_year': fy_key,
                'key__exact': EntityStateModel.KEY_PURCHASE_ORDER
            }

            state_model_qs = EntityStateModel.objects.filter(**LOOKUP).select_related(
                'entity_model').select_for_update()
            state_model = state_model_qs.get()
            state_model.sequence = F('sequence') + 1
            state_model.save()
            state_model.refresh_from_db()
            return state_model
        except ObjectDoesNotExist:
            EntityModel = lazy_loader.get_entity_model()
            entity_model = EntityModel.objects.get(uuid__exact=self.entity_id)
            fy_key = entity_model.get_fy_for_date(dt=self.date_draft)

            LOOKUP = {
                'entity_model_id': entity_model.uuid,
                'entity_unit_id': None,
                'fiscal_year': fy_key,
                'key': EntityStateModel.KEY_PURCHASE_ORDER,
                'sequence': 1
            }
            state_model = EntityStateModel.objects.create(**LOOKUP)
            return state_model
        except IntegrityError as e:
            if raise_exception:
                raise e

    def generate_po_number(self, commit: bool = False) -> str:
        """
        Atomic Transaction. Generates the next PurchaseOrder document number available.

        Parameters
        ----------
        commit: bool
            Commits transaction into PurchaseOrderModel.

        Returns
        -------
        str
            A String, representing the generated or current PurchaseOrderModel instance Document Number.
        """
        if self.can_generate_po_number():
            with transaction.atomic(durable=True):

                state_model = None
                while not state_model:
                    state_model = self._get_next_state_model(raise_exception=False)

                seq = str(state_model.sequence).zfill(DJANGO_LEDGER_DOCUMENT_NUMBER_PADDING)
                self.po_number = f'{DJANGO_LEDGER_PO_NUMBER_PREFIX}-{state_model.fiscal_year}-{seq}'

                if commit:
                    self.save(update_fields=['po_number'])

        return self.po_number

    def clean(self):
        if self.can_generate_po_number():
            self.generate_po_number(commit=True)


class PurchaseOrderModel(PurchaseOrderModelAbstract):
    """
    Purchase Order Base Model
    """

    class Meta(PurchaseOrderModelAbstract.Meta):
        abstract = False


def purchaseordermodel_presave(instance: PurchaseOrderModel, **kwargs):
    if instance.can_generate_po_number():
        instance.generate_po_number(commit=False)


pre_save.connect(receiver=purchaseordermodel_presave, sender=PurchaseOrderModel)



from django.urls import path

from django_ledger import views

urlpatterns = [
    # Entity Views ----
    path('list/', views.EntityModelListView.as_view(), name='entity-list'),
    path('create/', views.EntityModelCreateView.as_view(), name='entity-create'),

    # DASHBOARD Views...
    path('<slug:entity_slug>/dashboard/',
         views.EntityModelDetailHandlerView.as_view(),
         name='entity-dashboard'),
    path('<slug:entity_slug>/dashboard/year/<int:year>/',
         views.FiscalYearEntityModelDashboardView.as_view(),
         name='entity-dashboard-year'),
    path('<slug:entity_slug>/dashboard/quarter/<int:year>/<int:quarter>/',
         views.QuarterlyEntityDashboardView.as_view(),
         name='entity-dashboard-quarter'),
    path('<slug:entity_slug>/dashboard/month/<int:year>/<int:month>/',
         views.MonthlyEntityDashboardView.as_view(),
         name='entity-dashboard-month'),
    path('<slug:entity_slug>/dashboard/date/<int:year>/<int:month>/<int:day>/',
         views.DateEntityDashboardView.as_view(),
         name='entity-dashboard-date'),

    path('<slug:entity_slug>/update/', views.EntityModelUpdateView.as_view(), name='entity-update'),
    path('<slug:entity_slug>/delete/', views.EntityDeleteView.as_view(), name='entity-delete'),
]



"""
Django Ledger created by Miguel Sanda <msanda@arrobalytics.com>.
Copyright© EDMA Group Inc licensed under the GPLv3 Agreement.

Contributions to this module:
Miguel Sanda <msanda@arrobalytics.com>
"""

from django.forms import (ModelForm, DateInput, TextInput, Select, BaseModelFormSet,
                          modelformset_factory, Textarea, BooleanField, ValidationError)
from django.utils.translation import gettext_lazy as _

from django_ledger.models import (ItemModel, PurchaseOrderModel, ItemTransactionModel, EntityUnitModel)
from django_ledger.settings import DJANGO_LEDGER_FORM_INPUT_CLASSES


class PurchaseOrderModelCreateForm(ModelForm):
    def __init__(self, *args, entity_slug, user_model, **kwargs):
        super().__init__(*args, **kwargs)
        self.ENTITY_SLUG = entity_slug
        self.USER_MODEL = user_model

    class Meta:
        model = PurchaseOrderModel
        fields = [
            'po_title',
        ]
        widgets = {
            'po_title': TextInput(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES + ' is-large',
                'placeholder': 'What this PO is about...'})
        }
        labels = {
            'for_inventory': _('Is this an inventory purchase?')
        }


class BasePurchaseOrderModelUpdateForm(ModelForm):

    def __init__(self, *args, entity_slug, user_model, **kwargs):
        super().__init__(*args, **kwargs)
        self.ENTITY_SLUG = entity_slug
        self.USER_MODEL = user_model
        self.PO_MODEL: PurchaseOrderModel = self.instance

    class Meta:
        model = PurchaseOrderModel
        fields = [
            'markdown_notes'
        ]
        widgets = {
            'po_title': TextInput(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES + ' is-large'
            }),
            'po_status': Select(attrs={'class': DJANGO_LEDGER_FORM_INPUT_CLASSES}),
            'vendor': Select(attrs={'class': DJANGO_LEDGER_FORM_INPUT_CLASSES}),
            'date_fulfilled': DateInput(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES,
                'placeholder': _('Fulfillment Date (YYYY-MM-DD)...')
            }),
            'markdown_notes': Textarea(attrs={
                'class': 'textarea'
            })
        }
        labels = {
            'po_status': _('PO Status'),
            'fulfilled': _('Mark as Fulfilled'),
            'markdown_notes': _('PO Notes'),
        }


class DraftPurchaseOrderModelUpdateForm(BasePurchaseOrderModelUpdateForm):
    class Meta(BasePurchaseOrderModelUpdateForm.Meta):
        fields = [
            'po_title',
            'markdown_notes',
        ]


class ReviewPurchaseOrderModelUpdateForm(BasePurchaseOrderModelUpdateForm):
    class Meta(BasePurchaseOrderModelUpdateForm.Meta):
        fields = [
            'markdown_notes',
        ]


class ApprovedPurchaseOrderModelUpdateForm(BasePurchaseOrderModelUpdateForm):
    class Meta(BasePurchaseOrderModelUpdateForm.Meta):
        fields = [
            'markdown_notes',
        ]


class PurchaseOrderItemTransactionForm(ModelForm):
    create_bill = BooleanField(required=False)

    class Meta:
        model = ItemTransactionModel
        fields = [
            'item_model',
            'po_unit_cost',
            'po_quantity',
            'entity_unit',
            'po_item_status',
            'create_bill',
        ]
        widgets = {
            'item_model': Select(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES + ' is-small',
            }),
            'entity_unit': Select(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES + ' is-small',
            }),
            'po_item_status': Select(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES + ' is-small',
            }),
            'po_unit_cost': TextInput(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES + ' is-small',
            }),
            'po_quantity': TextInput(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES + ' is-small',
            })
        }

    def clean(self):
        cleaned_data = super(PurchaseOrderItemTransactionForm, self).clean()
        po_item_status = cleaned_data['po_item_status']
        po_item_model: ItemTransactionModel = self.instance
        if 'po_item_status' in self.changed_data:
            po_model: PurchaseOrderModel = getattr(self, 'PO_MODEL')
            if po_model.po_status == po_model.PO_STATUS_APPROVED:
                if not po_item_status:
                    raise ValidationError('Cannot assign null status to approved PO.')
                if all([
                    self.instance.bill_model_id,
                    po_item_status == ItemTransactionModel.STATUS_NOT_ORDERED
                ]):
                    raise ValidationError('Cannot assign not ordered status to a billed item. '
                                          'Void or delete bill first')
            if all([
                po_item_status in [ItemTransactionModel.STATUS_IN_TRANSIT, ItemTransactionModel.STATUS_RECEIVED],
                not po_item_model.bill_model_id
            ]):
                raise ValidationError(f'Cannot mark as {po_item_status.upper()}. '
                                      'Item must be billed first.')
        return cleaned_data


class BasePurchaseOrderItemFormset(BaseModelFormSet):

    def __init__(self, *args, entity_slug, user_model, po_model: PurchaseOrderModel, **kwargs):
        super().__init__(*args, **kwargs)
        self.USER_MODEL = user_model
        self.ENTITY_SLUG = entity_slug
        self.PO_MODEL = po_model

        items_qs = ItemModel.objects.for_po(
            entity_model=self.ENTITY_SLUG,
        )

        unit_qs = EntityUnitModel.objects.for_entity(
            entity_model=self.ENTITY_SLUG
        )

        for form in self.forms:
            form.PO_MODEL = self.PO_MODEL
            form.fields['item_model'].queryset = items_qs
            form.fields['entity_unit'].queryset = unit_qs
            if not self.PO_MODEL.can_edit_items():
                form.fields['po_unit_cost'].disabled = True
                form.fields['po_quantity'].disabled = True
                form.fields['entity_unit'].disabled = True
                form.fields['item_model'].disabled = True
                if not self.PO_MODEL.is_approved() or self.PO_MODEL.is_fulfilled():
                    form.fields['po_item_status'].disabled = True
                    form.fields['po_item_status'].widget.attrs['class'] += form.instance.get_status_css_class()
            # PO is Draft
            else:
                form.fields['po_item_status'].disabled = True

    def get_queryset(self):
        po_item_queryset, _ = self.PO_MODEL.get_itemtxs_data()
        return po_item_queryset


CanEditPurchaseOrderItemFormset = modelformset_factory(
    model=ItemTransactionModel,
    form=PurchaseOrderItemTransactionForm,
    formset=BasePurchaseOrderItemFormset,
    can_delete=True,
    extra=5
)

ReadOnlyPurchaseOrderItemFormset = modelformset_factory(
    model=ItemTransactionModel,
    form=PurchaseOrderItemTransactionForm,
    formset=BasePurchaseOrderItemFormset,
    can_delete=False,
    extra=0
)


def get_po_itemtxs_formset_class(po_model: PurchaseOrderModel):
    if po_model.is_draft():
        return CanEditPurchaseOrderItemFormset
    return ReadOnlyPurchaseOrderItemFormset



from django.urls import path

from django_ledger import views

urlpatterns = [

    # BALANCE SHEET Reports...
    # Entities...
    path('entity/<slug:entity_slug>/balance-sheet/',
         views.BaseBalanceSheetRedirectView.as_view(),
         name='entity-bs'),
    path('entity/<slug:entity_slug>/balance-sheet/year/<int:year>/',
         views.FiscalYearBalanceSheetView.as_view(),
         name='entity-bs-year'),
    path('entity/<slug:entity_slug>/balance-sheet/quarter/<int:year>/<int:quarter>/',
         views.QuarterlyBalanceSheetView.as_view(),
         name='entity-bs-quarter'),
    path('entity/<slug:entity_slug>/balance-sheet/month/<int:year>/<int:month>/',
         views.MonthlyBalanceSheetView.as_view(),
         name='entity-bs-month'),
    path('entity/<slug:entity_slug>/balance-sheet/date/<int:year>/<int:month>/<int:day>/',
         views.DateBalanceSheetView.as_view(),
         name='entity-bs-date'),

    # Ledgers....
    path('ledger/<slug:entity_slug>/<uuid:ledger_pk>/balance-sheet/',
         views.BaseLedgerModelBalanceSheetView.as_view(),
         name='ledger-bs'),
    path('ledger/<slug:entity_slug>/<uuid:ledger_pk>/balance-sheet/year/<int:year>/',
         views.FiscalYearLedgerModelBalanceSheetView.as_view(),
         name='ledger-bs-year'),
    path('ledger/<slug:entity_slug>/<uuid:ledger_pk>/balance-sheet/quarter/<int:year>/<int:quarter>/',
         views.QuarterlyLedgerModelBalanceSheetView.as_view(),
         name='ledger-bs-quarter'),
    path('ledger/<slug:entity_slug>/<uuid:ledger_pk>/balance-sheet/month/<int:year>/<int:month>/',
         views.MonthlyLedgerModelBalanceSheetView.as_view(),
         name='ledger-bs-month'),
    path('ledger/<slug:entity_slug>/<uuid:ledger_pk>/balance-sheet/date/<int:year>/<int:month>/<int:day>/',
         views.DateLedgerModelBalanceSheetView.as_view(),
         name='ledger-bs-date'),

    # Units...
    path('unit/<slug:entity_slug>/<slug:unit_slug>/balance-sheet/',
         views.BaseEntityUnitModelBalanceSheetView.as_view(),
         name='unit-bs'),
    path('unit/<slug:entity_slug>/<slug:unit_slug>/balance-sheet/year/<int:year>/',
         views.FiscalYearEntityUnitModelBalanceSheetView.as_view(),
         name='unit-bs-year'),
    path('unit/<slug:entity_slug>/<slug:unit_slug>/balance-sheet/quarter/<int:year>/<int:quarter>/',
         views.QuarterlyEntityUnitModelBalanceSheetView.as_view(),
         name='unit-bs-quarter'),
    path('unit/<slug:entity_slug>/<slug:unit_slug>/balance-sheet/month/<int:year>/<int:month>/',
         views.MonthlyEntityUnitModelBalanceSheetView.as_view(),
         name='unit-bs-month'),
    path('unit/<slug:entity_slug>/<slug:unit_slug>/balance-sheet/date/<int:year>/<int:month>/<int:day>/',
         views.DateEntityUnitModelBalanceSheetView.as_view(),
         name='unit-bs-date'),

    # INCOME STATEMENT Reports ----
    # Entity .....
    path('entity/<slug:entity_slug>/income-statement/',
         views.BaseIncomeStatementRedirectView.as_view(),
         name='entity-ic'),
    path('entity/<slug:entity_slug>/income-statement/year/<int:year>/',
         views.FiscalYearIncomeStatementView.as_view(),
         name='entity-ic-year'),
    path('entity/<slug:entity_slug>/income-statement/quarter/<int:year>/<int:quarter>/',
         views.QuarterlyIncomeStatementView.as_view(),
         name='entity-ic-quarter'),
    path('entity/<slug:entity_slug>/income-statement/month/<int:year>/<int:month>/',
         views.MonthlyIncomeStatementView.as_view(),
         name='entity-ic-month'),
    path('entity/<slug:entity_slug>/income-statement/date/<int:year>/<int:month>/<int:day>/',
         views.MonthlyIncomeStatementView.as_view(),
         name='entity-ic-date'),

    # Ledgers ....
    path('ledger/<slug:entity_slug>/<uuid:ledger_pk>/income-statement/',
         views.BaseLedgerIncomeStatementView.as_view(),
         name='ledger-ic'),
    path('ledger/<slug:entity_slug>/<uuid:ledger_pk>/income-statement/year/<int:year>/',
         views.FiscalYearLedgerIncomeStatementView.as_view(),
         name='ledger-ic-year'),
    path('ledger/<slug:entity_slug>/<uuid:ledger_pk>/income-statement/quarter/<int:year>/<int:quarter>/',
         views.QuarterlyLedgerIncomeStatementView.as_view(),
         name='ledger-ic-quarter'),
    path('ledger/<slug:entity_slug>/<uuid:ledger_pk>/income-statement/month/<int:year>/<int:month>/',
         views.MonthlyLedgerIncomeStatementView.as_view(),
         name='ledger-ic-month'),
    path('ledger/<slug:entity_slug>/<uuid:ledger_pk>/income-statement/date/<int:year>/<int:month>/<int:day>/',
         views.DateLedgerIncomeStatementView.as_view(),
         name='ledger-ic-date'),

    # Entity Units...
    path('unit/<slug:entity_slug>/<slug:unit_slug>/income-statement/',
         views.BaseEntityUnitModelIncomeStatementView.as_view(),
         name='unit-ic'),
    path('unit/<slug:entity_slug>/<slug:unit_slug>/income-statement/year/<int:year>/',
         views.FiscalYearEntityUnitModelIncomeStatementView.as_view(),
         name='unit-ic-year'),
    path('unit/<slug:entity_slug>/<slug:unit_slug>/income-statement/quarter/<int:year>/<int:quarter>/',
         views.QuarterlyIncomeStatementView.as_view(),
         name='unit-ic-quarter'),
    path('unit/<slug:entity_slug>/<slug:unit_slug>/income-statement/month/<int:year>/<str:month>/',
         views.MonthlyIncomeStatementView.as_view(),
         name='unit-ic-month'),
    path('unit/<slug:entity_slug>/<slug:unit_slug>/income-statement/date/<int:year>/<str:month>/<int:day>/',
         views.MonthlyIncomeStatementView.as_view(),
         name='unit-ic-date'),

    # CASH FLOW STATEMENTS...
    # Entities...
    path('entity/<slug:entity_slug>/cash-flow-statement/',
         views.BaseCashFlowStatementRedirectView.as_view(),
         name='entity-cf'),
    path('entity/<slug:entity_slug>/cash-flow-statement/year/<int:year>/',
         views.FiscalYearCashFlowStatementView.as_view(),
         name='entity-cf-year'),
    path('entity/<slug:entity_slug>/cash-flow-statement/quarter/<int:year>/<int:quarter>/',
         views.QuarterlyCashFlowStatementView.as_view(),
         name='entity-cf-quarter'),
    path('entity/<slug:entity_slug>/cash-flow-statement/month/<int:year>/<int:month>/',
         views.MonthlyCashFlowStatementView.as_view(),
         name='entity-cf-month'),
    path('entity/<slug:entity_slug>/cash-flow-statement/date/<int:year>/<int:month>/<int:day>/',
         views.DateCashFlowStatementView.as_view(),
         name='entity-cf-date'),

    # Ledgers...
    path('entity/<slug:entity_slug>/<uuid:ledger_pk>/cash-flow-statement/',
         views.BaseLedgerModelCashFlowStatementRedirectView.as_view(),
         name='ledger-cf'),
    path('entity/<slug:entity_slug>/<uuid:ledger_pk>/cash-flow-statement/year/<int:year>/',
         views.FiscalYearLedgerModelCashFlowStatementView.as_view(),
         name='ledger-cf-year'),
    path('entity/<slug:entity_slug>/<uuid:ledger_pk>/cash-flow-statement/quarter/<int:year>/<int:quarter>/',
         views.QuarterlyLedgerModelCashFlowStatementView.as_view(),
         name='ledger-cf-quarter'),
    path('entity/<slug:entity_slug>/<uuid:ledger_pk>/cash-flow-statement/month/<int:year>/<int:month>/',
         views.MonthlyLedgerModelCashFlowStatementView.as_view(),
         name='ledger-cf-month'),
    path('entity/<slug:entity_slug>/<uuid:ledger_pk>/cash-flow-statement/date/<int:year>/<int:month>/<int:day>/',
         views.DateLedgerModelCashFlowStatementView.as_view(),
         name='ledger-cf-date'),

    # Entity Units...
    path('unit/<slug:entity_slug>/<slug:unit_slug>/cash-flow-statement/',
         views.BaseEntityUnitModelCashFlowStatementView.as_view(),
         name='unit-cf'),
    path('unit/<slug:entity_slug>/<slug:unit_slug>/cash-flow-statement/year/<int:year>/',
         views.FiscalYearEntityUnitModelCashFlowStatementView.as_view(),
         name='unit-cf-year'),
    path('unit/<slug:entity_slug>/<slug:unit_slug>/cash-flow-statement/quarter/<int:year>/<int:quarter>/',
         views.QuarterlyEntityUnitModelCashFlowStatementView.as_view(),
         name='unit-cf-quarter'),
    path('unit/<slug:entity_slug>/<slug:unit_slug>/cash-flow-statement/month/<int:year>/<str:month>/',
         views.MonthlyEntityUnitModelCashFlowStatementView.as_view(),
         name='unit-cf-month'),
    path('unit/<slug:entity_slug>/<slug:unit_slug>/cash-flow-statement/date/<int:year>/<str:month>/<int:day>/',
         views.DateEntityUnitModelCashFlowStatementView.as_view(),
         name='unit-cf-date'),

]



# Generated by Django 5.1.6 on 2025-02-13 01:40

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('django_ledger', '0019_alter_transactionmodel_amount_and_more'),
    ]

    operations = [
        migrations.RemoveIndex(
            model_name='bankaccountmodel',
            name='django_ledg_cash_ac_59a8af_idx',
        ),
        migrations.RenameField(
            model_name='bankaccountmodel',
            old_name='cash_account',
            new_name='account_model',
        ),
        migrations.AlterUniqueTogether(
            name='bankaccountmodel',
            unique_together={('entity_model', 'account_model', 'account_number', 'routing_number'),
                             ('entity_model', 'account_number')},
        ),
        migrations.AlterField(
            model_name='bankaccountmodel',
            name='account_type',
            field=models.CharField(
                choices=[('checking', 'Checking'), ('savings', 'Savings'), ('credit_card', 'Credit Card'),
                         ('mortgage', 'Mortgage')], default='checking', max_length=20, verbose_name='Account Type'),
        ),
        migrations.AlterField(
            model_name='vendormodel',
            name='account_type',
            field=models.CharField(
                choices=[('checking', 'Checking'), ('savings', 'Savings'), ('credit_card', 'Credit Card'),
                         ('mortgage', 'Mortgage')], default='checking', max_length=20, verbose_name='Account Type'),
        ),
        migrations.AddIndex(
            model_name='bankaccountmodel',
            index=models.Index(fields=['account_model'], name='django_ledg_account_ed900d_idx'),
        ),
    ]



# Generated by Django 5.1.2 on 2024-11-20 22:26

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_ledger', '0017_alter_accountmodel_unique_together_and_more'),
    ]

    operations = [
        migrations.AddField(
            model_name='transactionmodel',
            name='cleared',
            field=models.BooleanField(default=False, verbose_name='Cleared'),
        ),
        migrations.AddField(
            model_name='transactionmodel',
            name='reconciled',
            field=models.BooleanField(default=False, verbose_name='Reconciled'),
        ),
        migrations.AlterField(
            model_name='chartofaccountmodel',
            name='entity',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='django_ledger.entitymodel', verbose_name='Entity'),
        ),
        migrations.AddIndex(
            model_name='transactionmodel',
            index=models.Index(fields=['cleared'], name='django_ledg_cleared_335c0b_idx'),
        ),
        migrations.AddIndex(
            model_name='transactionmodel',
            index=models.Index(fields=['reconciled'], name='django_ledg_reconci_75dbc7_idx'),
        ),
    ]



"""
Django Ledger created by Miguel Sanda <msanda@arrobalytics.com>.
Copyright© EDMA Group Inc licensed under the GPLv3 Agreement.

Contributions to this module:
    * Miguel Sanda <msanda@arrobalytics.com>
"""
from django.contrib import messages
from django.core.exceptions import ImproperlyConfigured, ValidationError
from django.urls import reverse
from django.utils.translation import gettext_lazy as _
from django.views.generic import ListView, CreateView, UpdateView, RedirectView
from django.views.generic.detail import SingleObjectMixin

from django_ledger.forms.bank_account import BankAccountCreateForm, BankAccountUpdateForm
from django_ledger.models import EntityModel
from django_ledger.models.bank_account import BankAccountModel
from django_ledger.views.mixins import DjangoLedgerSecurityMixIn


class BankAccountModelModelBaseView(DjangoLedgerSecurityMixIn):
    queryset = None

    def get_queryset(self):
        if self.queryset is None:
            entity_model: EntityModel = self.get_authorized_entity_instance()
            self.queryset = entity_model.bankaccountmodel_set.select_related('account_model', 'entity_model')
        return super().get_queryset()


class BankAccountModelListView(BankAccountModelModelBaseView, ListView):
    template_name = 'django_ledger/bank_account/bank_account_list.html'
    PAGE_TITLE = _('Bank Accounts')
    context_object_name = 'bank_accounts'
    extra_context = {
        'page_title': PAGE_TITLE,
        'header_title': PAGE_TITLE,
        'header_subtitle_icon': 'clarity:bank-line'
    }


class BankAccountModelCreateView(BankAccountModelModelBaseView, CreateView):
    template_name = 'django_ledger/bank_account/bank_account_create.html'
    PAGE_TITLE = _('Create Bank Account')
    extra_context = {
        'page_title': PAGE_TITLE,
        'header_title': PAGE_TITLE,
        'header_subtitle_icon': 'clarity:bank-line'
    }

    def get_form(self, form_class=None):
        return BankAccountCreateForm(
            entity_slug=self.kwargs['entity_slug'],
            user_model=self.request.user,
            **self.get_form_kwargs()
        )

    def get_success_url(self):
        return reverse('django_ledger:bank-account-list',
                       kwargs={
                           'entity_slug': self.kwargs['entity_slug']
                       })

    def form_valid(self, form):
        bank_account_model: BankAccountModel = form.save(commit=False)
        bank_account_model.configure(
            entity_slug=self.kwargs['entity_slug'],
            user_model=self.request.user,
            commit=False)
        return super(BankAccountModelCreateView, self).form_valid(form)


class BankAccountModelUpdateView(BankAccountModelModelBaseView, UpdateView):
    template_name = 'django_ledger/bank_account/bank_account_update.html'
    pk_url_kwarg = 'bank_account_pk'
    PAGE_TITLE = _('Update Bank Account')
    context_object_name = 'bank_account'
    extra_context = {
        'page_title': PAGE_TITLE,
        'header_title': PAGE_TITLE,
        'header_subtitle_icon': 'clarity:bank-line'
    }

    def get_success_url(self):
        return reverse('django_ledger:bank-account-list',
                       kwargs={
                           'entity_slug': self.kwargs['entity_slug']
                       })

    def get_form(self, form_class=None):
        return BankAccountUpdateForm(
            entity_slug=self.kwargs['entity_slug'],
            user_model=self.request.user,
            **self.get_form_kwargs()
        )


# ACTION VIEWS...
class BaseBankAccountModelActionView(BankAccountModelModelBaseView,
                                     RedirectView,
                                     SingleObjectMixin):
    http_method_names = ['get']
    pk_url_kwarg = 'bank_account_pk'
    action_name = None
    commit = True

    def get_redirect_url(self, *args, **kwargs):
        return reverse('django_ledger:bank-account-list',
                       kwargs={
                           'entity_slug': kwargs['entity_slug']
                       })

    def get(self, request, *args, **kwargs):
        kwargs['user_model'] = self.request.user
        if not self.action_name:
            raise ImproperlyConfigured('View attribute action_name is required.')
        response = super(BaseBankAccountModelActionView, self).get(request, *args, **kwargs)
        ba_model: BankAccountModel = self.get_object()

        try:
            getattr(ba_model, self.action_name)(commit=self.commit, **kwargs)
        except ValidationError as e:
            messages.add_message(request,
                                 message=e.message,
                                 level=messages.ERROR,
                                 extra_tags='is-danger')
        return response


class BankAccountModelActionMarkAsActiveView(BaseBankAccountModelActionView):
    action_name = 'mark_as_active'


class BankAccountModelActionMarkAsInactiveView(BaseBankAccountModelActionView):
    action_name = 'mark_as_inactive'



from random import choice
from urllib.parse import urlparse

from django.urls import reverse

from django_ledger.forms.account import AccountModelCreateForm
from django_ledger.io import roles, CREDIT
from django_ledger.models import EntityModel
from django_ledger.models.accounts import AccountModel
from django_ledger.tests.base import DjangoLedgerBaseTest
from django_ledger.urls.account import urlpatterns as account_urls


class AccountModelTests(DjangoLedgerBaseTest):
    N = 2

    def setUp(self):
        self.resolve_url_patterns(
            url_patterns=account_urls
        )

    def test_protected_views(self):

        self.logout_client()
        entity_model = self.get_random_entity_model()
        account_model: AccountModel = self.get_random_account(entity_model=entity_model)

        for path, kwargs in self.URL_PATTERNS.items():
            url_kwargs = dict()
            url_kwargs['entity_slug'] = entity_model.slug

            if 'coa_slug' in kwargs:
                url_kwargs['coa_slug'] = account_model.coa_slug
            if 'account_pk' in kwargs:
                url_kwargs['account_pk'] = account_model.uuid
            if 'year' in kwargs:
                url_kwargs['year'] = self.get_random_date().year
            if 'quarter' in kwargs:
                url_kwargs['quarter'] = choice(range(1, 5))
            if 'month' in kwargs:
                url_kwargs['month'] = self.get_random_date().month
            if 'day' in kwargs:
                url_kwargs['day'] = choice(range(1, 29))

            url = reverse(f'django_ledger:{path}', kwargs=url_kwargs)
            response = self.CLIENT.get(url, follow=False)
            redirect_url = urlparse(response.url)
            redirect_path = redirect_url.path
            login_path = reverse(viewname='django_ledger:login')

            self.assertEqual(response.status_code, 302, msg=f'{path} view is not protected.')
            self.assertEqual(redirect_path, login_path, msg=f'{path} view not redirecting to correct auth URL.')

    def test_account_create(self):

        entity_model = self.get_random_entity_model()
        account_create_url = reverse(
            viewname='django_ledger:account-create',
            kwargs={
                'entity_slug': entity_model.slug,
                'coa_slug': entity_model.default_coa_slug
            }
        )

        self.login_client()
        response = self.CLIENT.get(account_create_url)

        # check if user can access page...
        self.assertEqual(response.status_code, 200, msg="Fail to GET Account Create page.")

        # check if account create form is rendered...
        account_create_form: AccountModelCreateForm = response.context['form']
        self.assertContains(response, account_create_form.form_id, count=1)

        # check if all fields are rendered...
        self.assertContains(response, 'name="code"')
        self.assertContains(response, 'name="name"')
        self.assertContains(response, 'name="role"')
        self.assertContains(response, 'name="role_default"')
        self.assertContains(response, 'name="balance_type"')
        self.assertContains(response, 'name="active"')
        self.assertContains(response, 'name="coa_model"')

        # creates new account...
        NEW_ACCOUNT_CODE = '404000'
        form_data = {
            'code': NEW_ACCOUNT_CODE,
            'name': 'Test Income Account',
            'role': roles.INCOME_OPERATIONAL,
            'role_default': False,
            'balance_type': CREDIT,
            'active': True
        }
        response_create = self.CLIENT.post(account_create_url, data=form_data)
        self.assertEqual(response_create.status_code, 302)
        self.assertTrue(AccountModel.objects.for_entity(
            entity_model=entity_model,
            coa_model=entity_model.default_coa_slug,
        ).with_codes(codes=NEW_ACCOUNT_CODE).exists())

        # cannot create an account with same code again...
        response_create = self.CLIENT.post(account_create_url, data=form_data)
        self.assertEqual(response_create.status_code, 200)
        self.assertContains(response_create, 'Account with this Chart of Accounts and Account Code already exists')

    def test_account_activation(self):

        entity_model: EntityModel = self.get_random_entity_model()
        account_model: AccountModel = self.get_random_account(entity_model=entity_model, active=True)

        self.assertTrue(account_model.can_deactivate())
        self.assertTrue(account_model.active)

        account_model.deactivate(commit=True)
        self.assertFalse(account_model.can_deactivate())
        self.assertFalse(account_model.active)

        account_model.activate(commit=True)
        self.assertTrue(account_model.can_deactivate())
        self.assertTrue(account_model.active)

    def test_account_lock(self):
        entity_model: EntityModel = self.get_random_entity_model()
        account_model: AccountModel = self.get_random_account(entity_model=entity_model, active=True, locked=False)

        self.assertTrue(account_model.can_lock())
        self.assertFalse(account_model.can_unlock())

        account_model.lock(commit=True)
        self.assertFalse(account_model.can_lock())
        self.assertTrue(account_model.can_unlock())

        account_model.unlock(commit=True)
        self.assertTrue(account_model.can_lock())
        self.assertFalse(account_model.can_unlock())

    def test_annotations(self):
        entity_model: EntityModel = self.get_random_entity_model()
        account_model: AccountModel = self.get_random_account(entity_model=entity_model, active=True, locked=False)

        self.assertEqual(account_model.entity_slug, entity_model.slug)
        self.assertEqual(account_model.coa_slug, account_model.coa_model.slug)
        self.assertEqual(account_model.coa_model.active, account_model.is_coa_active())

    def test_can_transact(self):
        entity_model: EntityModel = self.get_random_entity_model()
        account_model: AccountModel = self.get_random_account(entity_model=entity_model, active=True, locked=False)
        self.assertTrue(account_model.can_transact())
        account_model.lock(commit=False)
        self.assertFalse(account_model.can_transact())
        account_model.unlock(commit=False)



"""
Django Ledger created by Miguel Sanda <msanda@arrobalytics.com>.
Copyright© EDMA Group Inc licensed under the GPLv3 Agreement.

Contributions to this module:
    * Miguel Sanda <msanda@arrobalytics.com>
"""
from collections import defaultdict
from datetime import datetime
from typing import Dict, Optional

from django.core.exceptions import ValidationError

from django_ledger.models.utils import lazy_loader


class IODigestValidationError(ValidationError):
    pass


class IODigestContextManager:

    def __init__(self, io_state: defaultdict):
        self.IO_DATA: Dict = io_state
        self.IO_RESULT = io_state['io_result']
        self.IO_MODEL = io_state['io_model']
        self.STRFTIME_FORMAT = '%B %d, %Y'

    def get_io_data(self) -> Dict:
        return self.IO_DATA

    def get_io_result(self):
        return self.IO_RESULT

    def get_io_txs_queryset(self):
        return self.IO_RESULT.txs_queryset

    def get_strftime_format(self):
        return self.STRFTIME_FORMAT

    @property
    def from_datetime(self):
        return self.get_from_datetime()

    def get_from_datetime(self, as_str: bool = False, fmt=None) -> Optional[datetime]:
        from_date = self.IO_DATA['from_date']
        if from_date:
            if as_str:
                if not fmt:
                    fmt = self.get_strftime_format()
                return from_date.strftime(fmt)
            return from_date

    @property
    def to_datetime(self):
        return self.get_to_datetime()

    def get_to_datetime(self, as_str: bool = False, fmt=None) -> datetime:
        if as_str:
            if not fmt:
                fmt = self.get_strftime_format()
            return self.IO_DATA['to_date'].strftime(fmt)
        return self.IO_DATA['to_date']

    def is_entity_model(self) -> bool:
        return isinstance(
            self.IO_MODEL,
            lazy_loader.get_entity_model()
        )

    def is_ledger_model(self) -> bool:
        return isinstance(
            self.IO_MODEL,
            lazy_loader.get_ledger_model()
        )

    def is_unit_model(self) -> bool:
        return isinstance(
            self.IO_MODEL,
            lazy_loader.get_entity_unit_model()
        )

    def is_by_unit(self) -> bool:
        return self.IO_DATA['by_unit']

    def is_by_period(self) -> bool:
        return self.IO_DATA['by_period']

    def is_by_activity(self) -> bool:
        return self.IO_DATA['by_activity']

    # Account Information
    def get_account_data(self, key_func=None) -> Dict:
        if key_func:
            return {
                key_func(acc): acc for acc in self.IO_DATA['accounts']
            }
        return {
            acc['account_uuid']: acc for acc in self.IO_DATA['accounts']
        }

    # Balance Sheet Data...
    def has_balance_sheet(self) -> bool:
        return 'balance_sheet' in self.IO_DATA

    def get_balance_sheet_data(self, raise_exception: bool = True) -> Dict:
        try:
            return self.IO_DATA['balance_sheet']
        except KeyError:
            if raise_exception:
                raise IODigestValidationError(
                    'IO Digest does not have balance sheet information available.'
                )

    # Income Statement Data...
    def has_income_statement(self) -> bool:
        return 'income_statement' in self.IO_DATA

    def get_income_statement_data(self, raise_exception: bool = True) -> Dict:
        try:
            return self.IO_DATA['income_statement']
        except KeyError:
            if raise_exception:
                raise IODigestValidationError(
                    'IO Digest does not have income statement information available.'
                )

    # Cash Flow Statement Data...
    def has_cash_flow_statement(self):
        return 'cash_flow_statement' in self.IO_DATA

    def get_cash_flow_statement_data(self, raise_exception: bool = True) -> Dict:
        try:
            return self.IO_DATA['cash_flow_statement']
        except KeyError:
            if raise_exception:
                raise IODigestValidationError(
                    'IO Digest does not have cash flow statement information available.'
                )

    # All Available Statements
    def get_financial_statements_data(self) -> Dict:
        return {
            'balance_sheet': self.get_balance_sheet_data() if self.has_balance_sheet() else None,
            'income_statement': self.get_income_statement_data() if self.has_income_statement() else None,
            'cash_flow_statement': self.get_cash_flow_statement_data() if self.has_cash_flow_statement() else None,
        }

    # CLOSING ENTRIES...

    def get_closing_entry_data(self):
        io_data = self.get_io_data()
        return io_data['accounts']



"""
Django Ledger created by Miguel Sanda <msanda@arrobalytics.com>.
Copyright© EDMA Group Inc licensed under the GPLv3 Agreement.

The LedgerModel is the heart of Django Ledger. It is a self-contained unit of accounting that implements a
double-entry accounting system capable of creating and managing transactions into the ledger and producing any financial
statements. In essence, an EntityModel is made of a collection of LedgerModels that drive the whole bookkeeping process.
Each LedgerModel is independent and they can operate as an individual or as a group.

Each LedgerModel encapsulates a collection of JournalEntryModels, which in turn hold a collection of TransactionModels.
LedgerModels can be used to represent any part of the EntityModel and can be extended to add additional functionality
and custom logic that drives how transactions are recorded into the books. One example of this is the LedgerWrapperMixIn
(see django_ledger.models.mixins.LedgerWrapperMixIn), which is the foundation of LedgerModel abstractions such as the
BillModel, InvoiceModel, PurchaseOrderModel and EstimateModel. Extending the LedgerModel can add additional
functionality necessary to implement industry-specific functionality to almost anything you can think of. Examples:
Farming Equipment, Real Estate, Investment Portfolio, etc.

Also, the LedgerModel inherits functionality from the all mighty IOMixIn (see django_ledger.io.io_mixin.IOMixIn),
which is the class responsible for making accounting queries to the Database in an efficient and performing way.
The digest() method executes all necessary aggregations and optimizations in order to push as much work to the Database
layer as possible in order to minimize the amount of data being pulled for analysis into the Python memory.

The Django Ledger core model follows the following structure:
EntityModel -< LedgerModel -< JournalEntryModel -< TransactionModel
"""
import warnings
from datetime import date
from string import ascii_lowercase, digits
from typing import Optional
from uuid import uuid4, UUID

from django.core.exceptions import ValidationError, ObjectDoesNotExist
from django.core.serializers.json import DjangoJSONEncoder
from django.db import models
from django.db.models import Q, Min, F, Count, Manager, QuerySet
from django.urls import reverse
from django.utils.translation import gettext_lazy as _

from django_ledger.io.io_core import IOMixIn
from django_ledger.models import lazy_loader
from django_ledger.models.deprecations import deprecated_entity_slug_behavior
from django_ledger.models.mixins import CreateUpdateMixIn
from django_ledger.models.signals import (
    ledger_posted,
    ledger_unposted,
    ledger_locked,
    ledger_unlocked,
    ledger_hidden,
    ledger_unhidden
)
from django_ledger.settings import DJANGO_LEDGER_USE_DEPRECATED_BEHAVIOR

LEDGER_ID_CHARS = ascii_lowercase + digits


class LedgerModelValidationError(ValidationError):
    pass


class LedgerModelQuerySet(QuerySet):
    """
    Custom QuerySet for filtering LedgerModel instances based on specific fields.

    Provides predefined filtering methods to simplify working with ledger data.
    These filters allow querying for entries that are locked, unlocked, posted,
    unposted, hidden, visible, or current.

    Methods
    -------
    locked()
        Filters instances based on the 'locked' attribute set to `True`.
    unlocked()
        Filters instances based on the 'locked' attribute set to `False`.
    posted()
        Filters a queryset to include only posted entries with 'posted' set to `True`.
    unposted()
        Filters a queryset to include only unposted entries with 'posted' set to `False`.
    hidden()
        Filters a queryset to include only items marked as hidden.
    visible()
        Filters out hidden items from the queryset.
    current()
        Filters the queryset to include items where the earliest timestamp
        comes after the entity's last closing date or is null.
    """

    def for_user(self, user_model) -> 'LedgerModelQuerySet':
        if user_model.is_superuser:
            return self
        return self.filter(
            Q(entity__admin=user_model) |
            Q(entity__managers__in=[user_model])
        )

    def locked(self) -> 'LedgerModelQuerySet':
        """
        Filters instances based on the 'locked' attribute.

        This method returns a new instance where the 'locked' attribute is set to `True`.

        Returns
        -------
        LedgerModelQuerySet
            A new instance filtered with the attribute 'locked' set to `True`.
        """
        return self.filter(locked=True)

    def unlocked(self) -> 'LedgerModelQuerySet':
        """
        Filters and returns an instance where the `locked` attribute is set to `False`.

        Returns
        -------
        LedgerModelQuerySet
            A new or modified instance of the object with the filter applied.
        """
        return self.filter(locked=False)

    def posted(self) -> 'LedgerModelQuerySet':
        """
        Filters a queryset to include only posted entries.

        Returns
        -------
        LedgerModelQuerySet
            A new queryset instance containing only objects with
            the `posted` field set to True.
        """
        return self.filter(posted=True)

    def unposted(self) -> 'LedgerModelQuerySet':
        """
        Filters a queryset or a similar iterable-like object to include only items that
        have the attribute `posted` set to `True`. This method returns a new instance
        containing the filtered results.

        Returns
        -------
        LedgerModelQuerySet
            A new instance containing only the filtered results with `posted` set to True.
        """
        return self.filter(posted=True)

    def hidden(self) -> 'LedgerModelQuerySet':
        """
        Filters the queryset to include only objects marked as hidden.

        Returns
        -------
        LedgerModelQuerySet
            A filtered queryset containing only objects with the `hidden` attribute
            set to True.
        """
        return self.filter(hidden=True)

    def visible(self) -> 'LedgerModelQuerySet':
        """
        Filters out hidden items from the queryset.

        Ensures that only items with the `hidden` attribute set to `False`
        are included in the returned queryset.

        Returns
        -------
        LedgerModelQuerySet
            A queryset containing only visible items.
        """
        return self.filter(hidden=False)

    def current(self) -> 'LedgerModelQuerySet':
        return self.filter(
            Q(earliest_timestamp__date__gt=F('entity__last_closing_date'))
            | Q(earliest_timestamp__isnull=True)
        )


class LedgerModelManager(Manager):
    """
    A custom-defined LedgerModelManager that implements custom QuerySet methods related to the LedgerModel.
    """

    def get_queryset(self) -> LedgerModelQuerySet:
        qs = LedgerModelQuerySet(self.model, using=self._db)
        return qs.select_related('entity').annotate(
            Count('journal_entries'),
            _entity_slug=F('entity__slug'),
            earliest_timestamp=Min('journal_entries__timestamp',
                                   filter=Q(journal_entries__posted=True)),
        )

    @deprecated_entity_slug_behavior
    def for_entity(self, entity_model: 'EntityModel | str | UUID' = None, **kwargs) -> LedgerModelQuerySet:
        """
        Returns a QuerySet of LedgerModels associated with a specific EntityModel & UserModel.
        May pass an instance of EntityModel or a String representing the EntityModel slug.

        Parameters
        ----------
        entity_model: str or EntityModel
            The entity slug or EntityModel used for filtering the QuerySet.

        Returns
        -------
        LedgerModelQuerySet
            A Filtered LedgerModelQuerySet.
        """
        EntityModel = lazy_loader.get_entity_model()

        qs = self.get_queryset()
        if 'user_model' in kwargs:
            warnings.warn(
                'user_model parameter is deprecated and will be removed in a future release. '
                'Use for_user(user_model).for_entity(entity_model) instead to keep current behavior.',
                DeprecationWarning,
                stacklevel=2
            )
            if DJANGO_LEDGER_USE_DEPRECATED_BEHAVIOR:
                qs = qs.for_user(kwargs['user_model'])

        if isinstance(entity_model, EntityModel):
            qs = qs.filter(entity=entity_model)
        elif isinstance(entity_model, str):
            qs = qs.filter(entity__slug__exact=entity_model)
        elif isinstance(entity_model, UUID):
            qs = qs.filter(entity_id=entity_model)
        else:
            raise LedgerModelValidationError(
                message='Must provide entity slug, EntityModel, or UUID parameter'
            )
        return qs


class LedgerModelAbstract(CreateUpdateMixIn, IOMixIn):
    """
    Base implementation of the LedgerModel.

    Attributes
    ----------
    uuid: UUID
        This is a unique primary key generated for the table. The default value of this field is uuid4().
    name: str
        Human-readable name of the LedgerModel. Maximum 150 characters.
    ledger_xid: str
        A unique user-defined identifier for the LedgerModel. Unique for the Entity Model.
    entity: EntityModel
        The EntityModel associated with the LedgerModel instance.
    posted: bool
        Determines if the LedgerModel is posted. Defaults to False. Mandatory.
    locked: bool
        Determines if the LedgerModel is locked. Defaults to False. Mandatory.
    hidden: bool
        Determines if the LedgerModel is hidden. Defaults to False. Mandatory.
    """
    _WRAPPED_MODEL_KEY = 'wrapped_model'
    uuid = models.UUIDField(default=uuid4, editable=False, primary_key=True)
    ledger_xid = models.SlugField(allow_unicode=True, max_length=150, null=True, blank=True,
                                  verbose_name=_('Ledger External ID'),
                                  help_text=_('User Defined Ledger ID'))
    name = models.CharField(max_length=150, null=True, blank=True, verbose_name=_('Ledger Name'))

    entity = models.ForeignKey('django_ledger.EntityModel',
                               on_delete=models.CASCADE,
                               verbose_name=_('Ledger Entity'))
    posted = models.BooleanField(default=False, verbose_name=_('Posted Ledger'))
    locked = models.BooleanField(default=False, verbose_name=_('Locked Ledger'))
    hidden = models.BooleanField(default=False, verbose_name=_('Hidden Ledger'))
    additional_info = models.JSONField(default=dict,
                                       encoder=DjangoJSONEncoder,
                                       null=True,
                                       blank=True)

    objects = LedgerModelManager.from_queryset(queryset_class=LedgerModelQuerySet)()

    class Meta:
        abstract = True
        ordering = ['-created']
        verbose_name = _('Ledger')
        verbose_name_plural = _('Ledgers')
        indexes = [
            models.Index(fields=['entity']),
            models.Index(fields=['entity', 'posted']),
            models.Index(fields=['entity', 'locked']),
            models.Index(fields=['entity', 'ledger_xid']),
            models.Index(fields=['ledger_xid']),
        ]
        unique_together = [
            ('entity', 'ledger_xid')
        ]

    def __str__(self):
        if self.name is not None:
            ledger_str = f'LedgerModel: {self.name}'
        elif self.ledger_xid is not None:
            ledger_str = f'LedgerModel: {self.ledger_xid}'
        else:
            ledger_str = f'LedgerModel: {self.uuid}'
        return f'{ledger_str} | Posted: {self.posted} | Locked: {self.locked}'

    @property
    def entity_slug(self):
        try:
            return getattr(self, '_entity_slug')
        except AttributeError:
            pass
        return self.entity.slug

    def has_wrapped_model_info(self):
        if self.additional_info is not None:
            return self._WRAPPED_MODEL_KEY in self.additional_info
        return False

    def has_wrapped_model(self):
        if self.has_wrapped_model_info():
            return True

        wrapped_model_info = self.get_wrapper_info
        for model_class, model_id in wrapped_model_info.items():
            try:
                return getattr(self, model_id)
            except ObjectDoesNotExist:
                pass
        return False

    def remove_wrapped_model_info(self):
        if self.has_wrapped_model_info():
            del self.additional_info[self._WRAPPED_MODEL_KEY]

    def has_jes_in_locked_period(self, force_evaluation: bool = True) -> bool:
        try:
            earliest_posted_je_timestamp = getattr(self, 'earliest_timestamp')
        except AttributeError:
            if force_evaluation:
                try:
                    earliest_je = self.journal_entries.posted().order_by('-timestamp').only('timestamp').first()
                    self.earliest_timestamp = earliest_je.timestamp if earliest_je else None
                except ObjectDoesNotExist:
                    self.earliest_timestamp = None
                earliest_posted_je_timestamp = getattr(self, 'earliest_timestamp')
            else:
                raise LedgerModelValidationError(
                    message=_(f'earliest_timestamp not present in LedgerModel {self.uuid}'))

        last_closing_date = self.get_entity_last_closing_date()
        if last_closing_date is None:
            return False
        if earliest_posted_je_timestamp is not None:
            earliest_posted_je_date = earliest_posted_je_timestamp.date()
            return earliest_posted_je_date <= last_closing_date
        return False

    def configure_for_wrapper_model(self, model_instance, commit: bool = False):

        if self.additional_info is None:
            self.additional_info = dict()

        wrapper_info = self.get_wrapper_info
        self.additional_info[self._WRAPPED_MODEL_KEY] = {
            'model': wrapper_info[model_instance.__class__],
            'uuid': model_instance.uuid
        }

        if commit:
            self.save(update_fields=[
                'additional_info',
                'updated'
            ])

    @property
    def get_wrapper_info(self):
        return {
            lazy_loader.get_bill_model(): 'billmodel',
            lazy_loader.get_invoice_model(): 'invoicemodel',
        }

    def get_wrapped_model_instance(self):
        if self.has_wrapped_model_info():
            return getattr(self, self.additional_info[self._WRAPPED_MODEL_KEY]['model'])

        for model_class, attr in self.get_wrapper_info.items():
            if getattr(self, attr, None):
                return getattr(self, attr)

    def get_wrapped_model_url(self):
        if self.has_wrapped_model():
            wrapped_model = self.get_wrapped_model_instance()
            return wrapped_model.get_absolute_url()

    def is_posted(self) -> bool:
        """
        Determines if the LedgerModel instance is posted.

        Returns
        -------
        bool
            True if posted, else False.
        """
        return self.posted is True

    def is_locked(self) -> bool:
        """
        Determines if the LedgerModel instance is locked.

        Returns
        -------
        bool
            True if locked, else False.
        """
        return self.locked is True

    def is_hidden(self) -> bool:
        """
        Determines if the LedgerModel instance is hidden.

        Returns
        -------
        bool
            True if hidden, else False.
        """
        return self.hidden is True

    def can_post(self) -> bool:
        """
        Determines if the LedgerModel can be marked as posted.

        Returns
        -------
        bool
            True if can be posted, else False.
        """
        return all([
            not self.is_posted(),
            not self.is_locked(),
            not self.has_jes_in_locked_period()
        ])

    def can_unpost(self) -> bool:
        """
        Determines if the LedgerModel can be un-posted.

        Returns
        -------
        bool
            True if can be un-posted, else False.
        """
        return all([
            self.is_posted(),
            not self.is_locked(),
            not self.has_jes_in_locked_period()
        ])

    def can_lock(self) -> bool:
        """
        Determines if the LedgerModel can be locked.

        Returns
        -------
        bool
            True if can be locked, else False.
        """
        return all([
            not self.is_locked(),
            self.is_posted()
        ])

    def can_unlock(self) -> bool:
        """
        Determines if the LedgerModel can be un-locked.

        Returns
        -------
        bool
            True if can be un-locked, else False.
        """
        return all([
            self.is_locked(),
            self.is_posted()
        ])

    def can_hide(self) -> bool:
        """
        Determines if the LedgerModel can be hidden.

        Returns
        -------
        bool
            True if can be hidden, else False.
        """
        return self.hidden is False

    def can_unhide(self) -> bool:
        """
        Determines if the LedgerModel can be un-hidden.

        Returns
        -------
        bool
            True if can be un-hidden, else False.
        """
        return self.hidden is True

    def can_delete(self) -> bool:
        if all([
            not self.is_locked(),
            not self.is_posted(),
            not self.has_wrapped_model_info(),
            not self.has_jes_in_locked_period()
        ]):
            return True
        return False

    def post(self, commit: bool = False, raise_exception: bool = True, **kwargs):
        """
        Posts the LedgerModel.

        Parameters
        ----------
        commit: bool
            If True, saves the LedgerModel instance instantly. Defaults to False.
        raise_exception:bool
            Raises LedgerModelValidationError if posting not allowed.
        """
        if not self.can_post():
            if raise_exception:
                raise LedgerModelValidationError(
                    message=_(f'Ledger {self.name} cannot be posted. UUID: {self.uuid}')
                )
            return
        self.posted = True
        if commit:
            self.save(update_fields=[
                'posted',
                'updated'
            ])
        ledger_posted.send_robust(sender=self.__class__,
                                  instance=self,
                                  commited=commit,
                                  **kwargs)

    def post_journal_entries(self, commit: bool = True, **kwargs):
        je_model_qs = self.journal_entries.unposted()
        for je_model in je_model_qs:
            je_model.mark_as_posted(raise_exception=False, commit=False)
        if commit:
            je_model_qs.bulk_update(objs=je_model_qs, fields=['posted', 'updated'])
        return je_model_qs

    def unpost(self, commit: bool = False, raise_exception: bool = True, **kwargs):
        """
        Un-posts the LedgerModel.

        Parameters
        ----------
        commit: bool
            If True, saves the LedgerModel instance instantly. Defaults to False.
        raise_exception:bool
            Raises LedgerModelValidationError if un-posting not allowed.
        """
        if not self.can_unpost():
            if raise_exception:
                raise LedgerModelValidationError(
                    message=_(f'Ledger {self.uuid} cannot be unposted.')
                )
            return
        self.posted = False
        if commit:
            self.save(update_fields=[
                'posted',
                'updated'
            ])
        ledger_unposted.send_robust(sender=self.__class__,
                                    instance=self,
                                    commited=commit,
                                    **kwargs)

    def lock(self, commit: bool = False, raise_exception: bool = True, **kwargs):
        """
        Locks the LedgerModel.

        Parameters
        ----------
        commit: bool
            If True, saves the LedgerModel instance instantly. Defaults to False.
        raise_exception: bool
            Raises LedgerModelValidationError if locking not allowed.
        """

        if not self.can_lock():
            if raise_exception:
                raise LedgerModelValidationError(
                    message=_(f'Ledger {self.name} cannot be locked. UUID: {self.uuid}')
                )
            return
        self.locked = True
        if commit:
            self.save(update_fields=[
                'locked',
                'updated'
            ])
        ledger_locked.send_robust(sender=self.__class__,
                                  instance=self,
                                  commited=commit,
                                  **kwargs)

    def lock_journal_entries(self, commit: bool = True, **kwargs):
        je_model_qs = self.journal_entries.unlocked()
        for je_model in je_model_qs:
            je_model.mark_as_locked(raise_exception=False, commit=False)
        if commit:
            je_model_qs.bulk_update(objs=je_model_qs, fields=['locked', 'updated'])
        return je_model_qs

    def unlock(self, commit: bool = False, raise_exception: bool = True, **kwargs):
        """
        Un-locks the LedgerModel.

        Parameters
        ----------
        commit: bool
            If True, saves the LedgerModel instance instantly. Defaults to False.
        """
        if not self.can_unlock():
            if raise_exception:
                raise LedgerModelValidationError(
                    message=_(f'Ledger {self.name} cannot be un-locked. UUID: {self.uuid}')
                )
            return

        self.locked = False
        if commit:
            self.save(update_fields=[
                'locked',
                'updated'
            ])
        ledger_unlocked.send_robust(sender=self.__class__,
                                    instance=self,
                                    commited=commit,
                                    **kwargs)

    def hide(self, commit: bool = False, raise_exception: bool = True, **kwargs):
        if not self.can_hide():
            if raise_exception:
                raise LedgerModelValidationError(
                    message=_(f'Ledger {self.name} cannot be hidden. UUID: {self.uuid}')
                )
            return
        self.hidden = True
        if commit:
            self.save(update_fields=[
                'hidden',
                'updated'
            ])
        ledger_hidden.send_robust(sender=self.__class__,
                                  instance=self,
                                  commited=commit,
                                  **kwargs)

    def unhide(self, commit: bool = False, raise_exception: bool = True, **kwargs):
        if not self.can_unhide():
            if raise_exception:
                raise LedgerModelValidationError(
                    message=_(f'Ledger {self.name} cannot be un-hidden. UUID: {self.uuid}')
                )
            return
        self.hidden = False
        if commit:
            self.save(update_fields=[
                'hidden',
                'updated'
            ])
        ledger_unhidden.send_robust(sender=self.__class__,
                                    instance=self,
                                    commited=commit,
                                    **kwargs)

    def delete(self, **kwargs):
        if not self.can_delete():
            raise LedgerModelValidationError(
                message=_(f'LedgerModel {self.name} cannot be deleted because posted is {self.is_posted()} '
                          f'and locked is {self.is_locked()}')
            )

        # checks if ledger model has journal entries in a closed period...
        if self.entity.has_closing_entry():
            earliest_je_timestamp = self.journal_entries.posted().order_by('-timestamp').values('timestamp').first()
            if earliest_je_timestamp is not None:
                earliest_date = earliest_je_timestamp['timestamp'].date()
                if earliest_date <= self.entity.last_closing_date:
                    raise LedgerModelValidationError(
                        message=_(
                            f'Journal Entries with date {earliest_date} cannot be deleted because of latest closing '
                            f'entry on {self.get_entity_last_closing_date()}')
                    )
        return super().delete(**kwargs)

    def get_entity_name(self) -> str:
        return self.entity.name

    def get_entity_last_closing_date(self) -> Optional[date]:
        return self.entity.last_closing_date

    # URLS....
    def get_absolute_url(self) -> str:
        """
        Determines the absolute URL of the LedgerModel instance.
        Results in additional Database query if entity field is not selected in QuerySet.

        Returns
        -------
        str
            URL as a string.
        """
        return reverse(viewname='django_ledger:je-list',
                       kwargs={
                           'entity_slug': self.entity.slug,
                           'ledger_pk': self.uuid
                       })

    def get_create_url(self) -> str:
        """
        Determines the update URL of the LedgerModel instance.
        Results in additional Database query if entity field is not selected in QuerySet.

        Returns
        -------
        str
            URL as a string.
        """
        return reverse('django_ledger:ledger-create',
                       kwargs={
                           'entity_slug': self.entity.slug
                       })

    def get_update_url(self) -> str:
        """
        Determines the update URL of the LedgerModel instance.
        Results in additional Database query if entity field is not selected in QuerySet.

        Returns
        -------
        str
            URL as a string.
        """
        return reverse('django_ledger:ledger-update',
                       kwargs={
                           'entity_slug': self.entity.slug,
                           'ledger_pk': self.uuid
                       })

    def get_list_url(self) -> str:
        """
        Determines the list URL of the LedgerModel instances.
        Results in additional Database query if entity field is not selected in QuerySet.

        Returns
        -------
        str
            URL as a string.
        """
        return reverse('django_ledger:ledger-list',
                       kwargs={
                           'entity_slug': self.entity.slug
                       })

    def get_journal_entry_list_url(self) -> str:
        return reverse('django_ledger:je-list',
                       kwargs={
                           'entity_slug': self.entity_slug,
                           'ledger_pk': self.uuid,
                       })

    def get_journal_entry_create_url(self) -> str:
        return reverse('django_ledger:je-create',
                       kwargs={
                           'entity_slug': self.entity_slug,
                           'ledger_pk': self.uuid
                       })

    def get_balance_sheet_url(self):
        return reverse(
            viewname='django_ledger:ledger-bs',
            kwargs={
                'entity_slug': self.entity.slug,
                'ledger_pk': self.uuid
            }
        )

    def get_income_statement_url(self):
        return reverse(
            viewname='django_ledger:ledger-ic',
            kwargs={
                'entity_slug': self.entity.slug,
                'ledger_pk': self.uuid
            }
        )

    def get_cash_flow_statement_url(self):
        return reverse(
            viewname='django_ledger:ledger-cf',
            kwargs={
                'entity_slug': self.entity.slug,
                'ledger_pk': self.uuid
            }
        )

    def get_delete_message(self):
        return _(f'Are you sure you want to delete Ledger {self.name} from Entity {self.get_entity_name()}?')

    # Action URL...
    def get_action_post_journal_entries_url(self):
        return reverse(
            viewname='django_ledger:ledger-action-post-journal-entries',
            kwargs={
                'entity_slug': self.entity_slug,
                'ledger_pk': self.uuid
            }
        )

    def get_action_lock_journal_entries_url(self):
        return reverse(
            viewname='django_ledger:ledger-action-lock-journal-entries',
            kwargs={
                'entity_slug': self.entity_slug,
                'ledger_pk': self.uuid
            }
        )


class LedgerModel(LedgerModelAbstract):
    """
    Base LedgerModel from Abstract.
    """

    class Meta(LedgerModelAbstract.Meta):
        abstract = False


def ledgermodel_presave(instance: LedgerModel, **kwargs):
    if not instance.has_wrapped_model_info():
        wrapper_instance = instance.get_wrapped_model_instance()
        instance.configure_for_wrapper_model(model_instance=wrapper_instance)



import warnings
from functools import wraps

from django_ledger.settings import DJANGO_LEDGER_USE_DEPRECATED_BEHAVIOR

if DJANGO_LEDGER_USE_DEPRECATED_BEHAVIOR:
    warnings.warn(
        message=(
            'You are using the deprecated behavior of django_ledger. '
            'Set DJANGO_LEDGER_USE_DEPRECATED_BEHAVIOR = False to transition to new API.'
        )
    )


def deprecated_entity_slug_behavior(func=None, *, message=None):
    """
    Decorator for for_entity(...) methods to warn about the deprecated `entity_slug` argument
    and optionally map it to `entity_model` for backward compatibility.

    Usage:
      @deprecated_for_entity_behavior
      def for_entity(self, entity_model=None, **kwargs):
          ...

      or with custom message:
      @deprecated_for_entity_behavior(message="custom deprecation message")
      def for_entity(...):
          ...
    """
    default_message = (
        'entity_slug parameter is deprecated and will be removed in a future release. '
        'Use entity_model instead (accepts EntityModel instance, UUID, or slug string).'
    )

    if func is None:
        # Called as @deprecated_for_entity_behavior(...)
        return lambda f: deprecated_entity_slug_behavior(f, message=message)

    @wraps(func)
    def wrapper(*args, **kwargs):
        if 'entity_slug' in kwargs and kwargs.get('entity_slug') is not None:
            warnings.warn(message or default_message, DeprecationWarning, stacklevel=2)

            # If both are provided, fail fast to avoid ambiguity
            if kwargs.get('entity_model') is not None:
                raise ValueError(
                    'Cannot specify both `entity_model` and `entity_slug`. '
                    'entity_slug is deprecated; pass only entity_model.'
                )

            # Maintain legacy behavior when enabled by settings toggle
            if DJANGO_LEDGER_USE_DEPRECATED_BEHAVIOR:
                kwargs['entity_model'] = kwargs.pop('entity_slug')
            else:
                # If deprecated behavior is disabled, drop the deprecated param
                # and rely on the method's own validation (likely to fail fast).
                kwargs.pop('entity_slug', None)

        return func(*args, **kwargs)

    return wrapper



from django.urls import path

from django_ledger import views

urlpatterns = [
    path('<slug:entity_slug>/latest/',
         views.PurchaseOrderModelListView.as_view(),
         name='po-list'),
    path('<slug:entity_slug>/year/<int:year>/',
         views.PurchaseOrderModelYearListView.as_view(),
         name='po-list-year'),
    path('<slug:entity_slug>/month/<int:year>/<int:month>/',
         views.PurchaseOrderModelMonthListView.as_view(),
         name='po-list-month'),
    path('<slug:entity_slug>/create/',
         views.PurchaseOrderModelCreateView.as_view(),
         name='po-create'),
    path('<slug:entity_slug>/create/estimate/<uuid:ce_pk>/',
         views.PurchaseOrderModelCreateView.as_view(for_estimate=True),
         name='po-create-estimate'),
    path('<slug:entity_slug>/detail/<uuid:po_pk>/',
         views.PurchaseOrderModelDetailView.as_view(),
         name='po-detail'),
    path('<slug:entity_slug>/update/<uuid:po_pk>/',
         views.PurchaseOrderModelUpdateView.as_view(),
         name='po-update'),
    path('<slug:entity_slug>/update/<uuid:po_pk>/update-items/',
         views.PurchaseOrderModelUpdateView.as_view(action_update_items=True),
         name='po-update-items'),
    path('<slug:entity_slug>/delete/<uuid:po_pk>/',
         views.PurchaseOrderModelDeleteView.as_view(),
         name='po-delete'),

    # Actions....
    path('<slug:entity_slug>/action/<uuid:po_pk>/mark-as-draft/',
         views.PurchaseOrderMarkAsDraftView.as_view(),
         name='po-action-mark-as-draft'),
    path('<slug:entity_slug>/action/<uuid:po_pk>/mark-as-review/',
         views.PurchaseOrderMarkAsReviewView.as_view(),
         name='po-action-mark-as-review'),
    path('<slug:entity_slug>/action/<uuid:po_pk>/mark-as-approved/',
         views.PurchaseOrderMarkAsApprovedView.as_view(),
         name='po-action-mark-as-approved'),
    path('<slug:entity_slug>/action/<uuid:po_pk>/mark-as-fulfilled/',
         views.PurchaseOrderMarkAsFulfilledView.as_view(),
         name='po-action-mark-as-fulfilled'),
    path('<slug:entity_slug>/action/<uuid:po_pk>/mark-as-canceled/',
         views.PurchaseOrderMarkAsCanceledView.as_view(),
         name='po-action-mark-as-canceled'),
    path('<slug:entity_slug>/action/<uuid:po_pk>/mark-as-void/',
         views.PurchaseOrderMarkAsVoidView.as_view(),
         name='po-action-mark-as-void'),
]



# Generated by Django 5.1.1 on 2024-10-09 19:40

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_ledger', '0016_remove_accountmodel_django_ledg_coa_mod_e19964_idx_and_more'),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name='accountmodel',
            unique_together=set(),
        ),
        migrations.AlterField(
            model_name='accountmodel',
            name='coa_model',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='django_ledger.chartofaccountmodel', verbose_name='Chart of Accounts'),
        ),
        migrations.AddConstraint(
            model_name='accountmodel',
            constraint=models.UniqueConstraint(fields=('coa_model', 'code'), name='unique_code_for_coa_model', violation_error_message='Account codes must be unique for each Chart of Accounts Model.'),
        ),
        migrations.AddConstraint(
            model_name='accountmodel',
            constraint=models.UniqueConstraint(fields=('coa_model', 'role', 'role_default'), name='only_one_account_assigned_as_default_for_role', violation_error_message='Only one default account for role permitted.'),
        ),
    ]



from random import choice
from urllib.parse import urlparse

from django.contrib.auth import get_user_model
from django.urls import reverse

from django_ledger.tests.base import DjangoLedgerBaseTest
from django_ledger.urls.closing_entry import urlpatterns as closing_entry_urls

UserModel = get_user_model()


class ClosingEntryModelTests(DjangoLedgerBaseTest):

    def setUp(self) -> None:
        self.URL_PATTERNS = {
            p.name: set(p.pattern.converters.keys()) for p in closing_entry_urls
        }

        self.MONTHS = list(range(1, 13))
        self.YEARS = [self.START_DATE.year, self.START_DATE.year + 1]

    def test_closing_entry_create(self):
        entity_model = self.get_random_entity_model()
        for y in self.YEARS:
            for m in self.MONTHS:
                entity_model.close_books_for_month(
                    year=y,
                    month=m,
                    post_closing_entry=True
                )

    def test_protected_views(self):
        self.logout_client()
        entity_model = self.get_random_entity_model()
        entity_model.close_entity_books(
            closing_date=self.get_random_date()
        )
        closing_entry_model = choice(entity_model.get_closing_entries())

        for path, kwargs in self.URL_PATTERNS.items():
            url_kwargs = dict()
            url_kwargs['entity_slug'] = entity_model.slug
            if 'year' in kwargs:
                url_kwargs['year'] = closing_entry_model.closing_date.year
            if 'month' in kwargs:
                url_kwargs['month'] = closing_entry_model.closing_date.month
            if 'closing_entry_pk' in kwargs:
                url_kwargs['closing_entry_pk'] = closing_entry_model.uuid

            url = reverse(f'django_ledger:{path}', kwargs=url_kwargs)
            response = self.CLIENT.get(url, follow=False)
            redirect_url = urlparse(response.url)
            redirect_path = redirect_url.path
            login_path = reverse(viewname='django_ledger:login')

            self.assertEqual(response.status_code, 302, msg=f'{path} view is not protected.')
            self.assertEqual(redirect_path, login_path, msg=f'{path} view not redirecting to correct auth URL.')

    def test_closing_entry_list(self):
        self.login_client()
        entity_model = self.get_random_entity_model()
        url = reverse('django_ledger:closing-entry-list', kwargs={'entity_slug': entity_model.slug})
        with self.assertNumQueries(4):
            response = self.CLIENT.get(path=url)



from datetime import date
from random import choice
from urllib.parse import urlparse

from django.conf import settings
from django.contrib.auth import get_user_model
from django.urls import reverse

from django_ledger.io.io_core import get_localdate
from django_ledger.models import EntityModel
from django_ledger.tests.base import DjangoLedgerBaseTest
from django_ledger.urls.entity import urlpatterns as entity_urls

UserModel = get_user_model()


class EntityModelTests(DjangoLedgerBaseTest):

    def setUp(self) -> None:
        super(EntityModelTests, self).setUp()

        self.ENTITY_URL_PATTERN = {
            p.name: set(p.pattern.converters.keys()) for p in entity_urls
        }

    # CLOSING ENTRY Tests...
    def test_closing_entry_dates_meta_key(self):
        self.logger.info('test_closing_entry_dates_meta_key...')
        for entity_model in self.ENTITY_MODEL_QUERYSET:
            self.assertTrue(isinstance(entity_model.meta, dict))
            self.assertTrue(entity_model.META_KEY_CLOSING_ENTRY_DATES in entity_model.meta)
            self.assertTrue(isinstance(entity_model.meta[entity_model.META_KEY_CLOSING_ENTRY_DATES], list))
            self.assertEqual(len(entity_model.meta[entity_model.META_KEY_CLOSING_ENTRY_DATES]), 0)

    def test_closing_entry_creation(self):
        for entity_model in self.ENTITY_MODEL_QUERYSET:
            start_year = self.START_DATE.year
            start_month = self.START_DATE.month
            while True:
                ce_list = entity_model.close_books_for_month(year=start_year, month=start_month)
                if all([
                    self.START_DATE.year + 1 == start_year,
                    self.START_DATE.month == start_month
                ]):
                    break
                elif start_month == 12:
                    start_month = 1
                    start_year += 1
                else:
                    start_month += 1

#     def test_closing_entry_meta(self):
#         self.logger.info('test_closing_entry_creation...')
#         for entity_model in self.ENTITY_MODEL_QUERYSET:

    # UI Tests....
    def test_ui_protected_views(self, test_date: date = None):
        """
        All Entity Model views must have user authenticated.
        If not, user must be redirected to login page.
        @param test_date: Optional test date. If None, will generate random date.
        """
        self.logout_client()
        # entity_model_qs = EntityModel.objects.for_user(self.user_model)

        for entity_model in self.ENTITY_MODEL_QUERYSET:
            for path, kwargs in self.ENTITY_URL_PATTERN.items():
                url_kwargs = dict()
                if 'entity_slug' in kwargs:
                    url_kwargs['entity_slug'] = entity_model.slug
                if 'year' in kwargs:
                    test_date = self.get_random_date()
                    url_kwargs['year'] = test_date.year
                if 'month' in kwargs:
                    url_kwargs['month'] = test_date.month
                if 'quarter' in kwargs:
                    url_kwargs['quarter'] = choice(range(1, 5))
                if 'day' in kwargs:
                    url_kwargs['day'] = test_date.day

                url = reverse(f'django_ledger:{path}', kwargs=url_kwargs)
                response = self.CLIENT.get(url, follow=False)
                redirect_url = urlparse(response.url)
                redirect_path = redirect_url.path
                self.assertEqual(response.status_code, 302, msg=f'{path} is not protected.')
                self.assertEqual(redirect_path, settings.LOGIN_URL, msg=f'{path} not redirecting to correct auth URL.')

    def test_ui_entity_create(self):
        """
        Testing the creation of a number of entities.
        """
        self.login_client()

        # ENTITY-CREATE VIEW...
        entity_create_url = reverse('django_ledger:entity-create')
        response = self.CLIENT.get(entity_create_url, follow=False)

        # making sure user is logged in...
        if response.status_code == 302:
            response = self.CLIENT.get(entity_create_url)
            self.assertContains(response, status_code=200, text='New Entity Information')
            self.assertContains(response, status_code=200, text='Populate Default CoA')
            self.assertContains(response, status_code=200, text='Activate All Accounts')
            self.assertContains(response, status_code=200, text='Fill With Sample Data?')
            self.assertContains(response, text='Create Entity')

        # checks that form throws Validation Error if any value is missing...
        # entity_must_have_all = ['city', 'state', 'zip_code', 'country']
        # for ent_data in self.TEST_DATA:
        #     while entity_must_have_all:
        #         ent_copy = ent_data.copy()
        #         del ent_copy[entity_must_have_all.pop()]
        #         response = self.CLIENT.post(entity_create_url, data=ent_copy, follow=False)
        #         self.assertContains(response, status_code=200, text='New Entity Information')
        #         self.assertFormError(response,
        #                              form='form',
        #                              field=None,
        #                              errors='Must provide all City/State/Zip/Country')

        # checks that valid url is provided...
        # ent_copy = self.get_random_entity_data()
        # ent_copy['website'] = ent_copy['website'][1:]
        # response = self.CLIENT.post(entity_create_url, data=ent_copy)
        # self.assertFormError(response, form='form', field='website', errors='Enter a valid URL.')

        # checks that a valid entity name is provided...
        # ent_copy = self.get_random_entity_data()
        # ent_copy['name'] = ''
        # response = self.CLIENT.post(entity_create_url, data=ent_copy, follow=False)
        # self.assertFormError(response, form='form', field='name', errors='Please provide a valid name for new Entity.')

        # checks for valid entity name length....
        # ent_copy = self.get_random_entity_data()
        # ent_copy['name'] = 'In'
        # response = self.CLIENT.post(entity_create_url, data=ent_copy, follow=False)
        # self.assertFormError(response, form='form', field='name', errors='Looks like this entity name is too short...')

        # creating a number of entities...
        for ent_data in self.TEST_DATA:
            response = self.CLIENT.post(entity_create_url, data=ent_data, follow=True)
            # user must be redirected if success...
            # self.assertContains(response, status_code=200, text='My Dashboard')
            # self.assertContains(response, status_code=200, text=ent_data['name'])

    def test_ui_entity_list(self):
        self.login_client()
        # self.refresh_test_data()

        # entity_models = self.get_entity_models()
        entity_list_url = reverse('django_ledger:entity-list')

        # ENTITY-LIST VIEW...
        with self.assertNumQueries(3):
            response = self.CLIENT.get(entity_list_url)

            # checks if it was able to render template...
            self.assertContains(response, status_code=200, text='My Entities')

            # checks if all entities where rendered...
            for ent_data in self.TEST_DATA:
                self.assertContains(response, status_code=200, text=ent_data['name'])

            # checks if all entities have proper anchor tags to dashboard and update views...
            entity_qs = response.context['entities']
            for entity_model in entity_qs:
                # checks if entity shows up in the list...
                self.assertContains(response,
                                    status_code=200,
                                    text=entity_model.name,
                                    msg_prefix=f'Entity {entity_model.name} not in the view!')

                # checks if there is a button with a link to the dashboard...
                self.assertContains(response,
                                    status_code=200,
                                    msg_prefix=f'There is no Dashboard link button for {entity_model.name}',
                                    text=entity_list_url)
                # checks if there is a button with a link to the delete view...
                self.assertContains(response,
                                    status_code=200,
                                    text=reverse('django_ledger:entity-delete',
                                                 kwargs={
                                                     'entity_slug': entity_model.slug
                                                 }))  # ENTITY-LIST VIEW...

    def test_ui_entity_update(self):

        self.login_client()
        entity_models = self.create_entity_models()
        entity_list_url = reverse('django_ledger:entity-list')
        an_entity: EntityModel = choice(self.ENTITY_MODEL_QUERYSET)

        # ENTITY-UPDATE VIEW...
        with self.assertNumQueries(3):
            entity_update_url = reverse('django_ledger:entity-update',
                                        kwargs={
                                            'entity_slug': an_entity.slug
                                        })
            response = self.CLIENT.get(entity_update_url)

        with self.assertNumQueries(4):  # previously 5
            ent_data = response.context['form'].initial
            ent_data['name'] = 'New Cool Name LLC'
            ent_data = {k: v for k, v in ent_data.items() if v}
            response = self.CLIENT.post(entity_update_url, data=ent_data)

        with self.assertNumQueries(3):
            # redirects to entity list
            self.assertRedirects(response, expected_url=entity_list_url)

        with self.assertNumQueries(3):
            response = self.CLIENT.get(entity_list_url)
            # checks if updated entity is in list...
            self.assertContains(response, status_code=200, text=ent_data['name'])

    def test_ui_entity_detail(self):

        self.login_client()
        entity_model: EntityModel = choice(self.ENTITY_MODEL_QUERYSET)

        # ENTITY-DETAIL VIEW...
        with self.assertNumQueries(2):
            # this will redirect to entity-detail-month...
            entity_detail_url = reverse('django_ledger:entity-dashboard',
                                        kwargs={
                                            'entity_slug': entity_model.slug
                                        })
            response = self.CLIENT.get(entity_detail_url)

        with self.assertNumQueries(8):  # previously 10
            local_dt = get_localdate()
            entity_month_detail_url = reverse('django_ledger:entity-dashboard-month',
                                              kwargs={
                                                  'entity_slug': entity_model.slug,
                                                  'year': local_dt.year,
                                                  'month': local_dt.month
                                              })
            self.assertRedirects(response, entity_month_detail_url)

        with self.assertNumQueries(8):
            # same as before, but this time the session must not be update because user has not suited entities...
            response = self.CLIENT.get(entity_month_detail_url)
            self.assertContains(response, text=entity_model.name)
            self.assertContains(response, text='Dashboard')
            self.assertTrue(response.context['bills'].count() >= 0)
            self.assertTrue(response.context['invoices'].count() >= 0)

    def test_ui_delete_entity(self):
        self.login_client()
        # entity_models = self.create_entity_models(n=1)
        entity_model = choice(self.ENTITY_MODEL_QUERYSET)
        # ENTITY-DELETE VIEW...
        with self.assertNumQueries(4):
            entity_delete_url = reverse('django_ledger:entity-delete',
                                        kwargs={
                                            'entity_slug': entity_model.slug
                                        })
            response = self.CLIENT.get(entity_delete_url)
            self.assertContains(response,
                                status_code=200,
                                text=entity_model.slug)
            self.assertContains(response,
                                status_code=200,
                                text=f'Are you sure you want to delete')
            self.assertContains(response,
                                status_code=200,
                                text=entity_delete_url)

        # this is a complex operation that requires several queries...
        response = self.CLIENT.post(entity_delete_url,
                                    data={
                                        'slug': entity_model.slug
                                    }, follow=False)

        with self.assertNumQueries(3):
            # checks that user is redirected to home after entity is deleted...
            home_url = reverse('django_ledger:home')
            self.assertRedirects(response, home_url)

        with self.assertNumQueries(3):
            # checks that entity no longer shows in entity list...
            home_url = reverse('django_ledger:home')
            response = self.CLIENT.get(home_url)
            self.assertNotContains(response, text=entity_model.slug)
